
#include "palmdep.h"
#include "endianutils.h"
#include "pnobridge.h"
#include "oscompat.h"
#include "endianutils.h"

#include "StarterRsc.h"
#include "palm5way.h"
#include "pnobridge.h"

#include "shared.h"

#include "cycloneoff.h"

#ifdef __CYCLONE__
typedef struct {uint32 __r9;uint32 __r10;} reg_sys_t;
static reg_sys_t reg_sys;
#endif

UInt32 stubzodiacblit ( UInt32 mode );
void PutImage ( void );
void audio_gen ( void );
unsigned char run_host ( void );
void stubaudiof ( void *s );
Err audio_callback_extended
  ( void *userData, SndStreamRef stream, void *buffer, UInt32* bufsize );
void *MemCart ( void );
void *MemCartFree ( void );
UInt32 host_get_skip ( UInt8 foo );

UInt8 *backbuffer = NULL;

// features
#define F_FPS 1           /* nolonger needed */
#define F_DO_EVENTS 1

#ifdef F_FPS
UInt32 fps_lasttime = 0;
UInt32 fps_counter = 0;
char fps_buffer [ 10 ] = "FPS: ";
#endif

// conveniance pointers
UInt16 display_pitch;
UInt16 ticks_per_sec;

// Linker still looks for ARMlet_Main as entry point, but the
// "ARMlet" name is now officially discouraged.  Compare an
// contrast to "PilotMain" for 68K applications.
#define PNO_Main ARMlet_Main

// Define DEBUG as 0 to turn off debugging support in the PNO.
// With this on, the PNO will register itself with the Palm OS
// Debugger on startup, allowing stepping through the code.
#define DEBUG 0

// ------------------------

unsigned long PNO_Main(
	const void *emulStateP, 
	void *userData68KP, 
	Call68KFuncType *call68KFuncP);

unsigned long Run ( void );

Call68KFuncType * g_my_call68KFuncP;
const void *g_my_emulStateP;
pnobridge_t *g_pnobridge;

UInt32 g_host = 0;
UInt8 g_running = 0;
UInt8 g_frameskip = 2;

unsigned long PNO_Main(
	const void *emulStateP, 
	void *userData68KP, 
	Call68KFuncType *call68KFuncP)
{

#ifdef PNODEBUG
  AdnDebugNativeRegister ( 'appl', 'STRT', 'ARMC', 1 );
#endif

  // needed before making any OS calls using the 
  // PACEInterface library
  InitPACEInterface ( emulStateP, call68KFuncP );

  g_my_call68KFuncP = call68KFuncP;
  g_my_emulStateP = emulStateP;
  g_pnobridge = userData68KP;

  // cheap hack of a bridge between app sides
  g_pnobridge -> vfs_max = ByteSwap32 ( g_pnobridge -> vfs_max );
  g_pnobridge -> vfs_vol = ByteSwap32 ( g_pnobridge -> vfs_vol );
  g_pnobridge -> vram = ByteSwap32 ( g_pnobridge -> vram );
  g_pnobridge -> eventf = ByteSwap32 ( g_pnobridge -> eventf );
  g_pnobridge -> display_width = ByteSwap32 ( g_pnobridge -> display_width );
  g_pnobridge -> display_height = ByteSwap32 ( g_pnobridge -> display_height );
  g_pnobridge -> display_pitch = ByteSwap32 ( g_pnobridge -> display_pitch );
  g_pnobridge -> game_rom = ByteSwap32 ( g_pnobridge -> game_rom );
  g_pnobridge -> audio_buf = ByteSwap32 ( g_pnobridge -> audio_buf );
  g_pnobridge -> flags = ByteSwap32 ( g_pnobridge -> flags );

  display_pitch = g_pnobridge -> display_pitch;

  // do something useful
  return ( Run() );
}

unsigned long Run ( void ) {
  extern unsigned char *cart_rom;   /* Cartridge ROM */
  UInt32 g_host = 0;
  UInt32 keys;
  UInt8 hostcount = 0;
  UInt8 frameskip = g_frameskip;
  extern UInt32 cart_rom_size;
  UInt8 nofps = 0;

  /* kickoff
   */
  ticks_per_sec = SysTicksPerSecond();

  /* init engine
   */

  /* init
   */
  debuglog ( "Load ROM ..." );
  if ( ! load_rom ( (char*) ( g_pnobridge -> game_rom ) ) ) {
    return ( 0 ); // failed
  }

  /* define display
   */
  if ( g_pnobridge -> flags & BF_FULLSCREEN ) {
    if ( g_pnobridge -> flags & BF_ZODIAC ) {
      // don't bother with backbuffer for zodiac; use GPU RAM instead
    } else {
      backbuffer = MemGluePtrNewL ( 640 * 480 * 2 );
      if ( ! backbuffer ) {
	DEBUGS ( "Couldn't get enough memory for scaling." );
      }
    }
  }

  memset ( &bitmap, sizeof(t_bitmap), 0 );
  bitmap.width  = 320;
  bitmap.height = 200;
  bitmap.depth  = 16;
  bitmap.granularity = 2; /* Bytes per pixel */

  if ( backbuffer ) {
    bitmap.pitch  = 640 * 2;
    bitmap.data   = (unsigned char *) backbuffer + ( bitmap.pitch * 32 );
  } else {

    if ( g_pnobridge -> flags & BF_FULLSCREEN &&
	 g_pnobridge -> flags & BF_ZODIAC )
    {
      bitmap.data = (unsigned char *) (UInt16*) stubzodiacblit ( 1 );
      bitmap.data -= 64;
      bitmap.pitch = g_pnobridge -> display_pitch * 2;
      nofps = 1;
    } else {
      bitmap.data   = (unsigned char *) g_pnobridge -> vram - 64 + ( display_pitch * 32 * 2 ); /* Pointer to your bitmap */
      bitmap.pitch = g_pnobridge -> display_pitch * 2; //(bitmap.width * bitmap.granularity);
    }

  }

  bitmap.viewport.w = 256; /* Initial size of display on power-up (do not change) */
  bitmap.viewport.h = 224;
  bitmap.viewport.x = 0x20; /* Offset used for rendering (do not change) */
  bitmap.viewport.y = 0x00;
  bitmap.remap = 1;

#ifdef __CYCLONE__
    __asm 
    {
        ldr  r0, = reg_sys
	add	 r0,r0,r10
	str	 r9,[r0]
	str	 r10,[r0,#4]
    } 
#endif

  /* start emu; includes emu main loop..
   */
  debuglog ( "Init the emulator" );
  system_init();

  /* fire up audio
   */
  if ( g_pnobridge -> flags & BF_AUDIO ) {
    extern UInt8 emul_z80;
    debuglog ( "Init audio" );
    audio_init ( AUDIO_RATE );
    emul_z80 = 1;

#ifdef AUDIO_ARM
    // set audio function
    {
      hostcall_t h;
      h.a = (UInt32) ByteSwap32(audio_callback_extended);
      h.b = (UInt32) &snd;
      //h.b = g_pnobridge -> audio_buf; // works
      stubaudiof ( &h );
    }
#endif

  } // audio

  /* reset
   */
  debuglog ( "Reset the Genesis" );
  system_reset();

  // zodiac keys
#define	keyBitJogBack		0x00008000	// jog wheel back button
#define keyBitFunction          keyBitJogBack
#define keyBitTriggerA          0x04000000
#define keyBitTriggerB          0x08000000
#define keyBitActionA           0x10000000
#define keyBitActionB           0x20000000
#define keyBitActionC           0x40000000
#define keyBitActionD           0x80000000

  /* run it!
   */
  g_running = 1;
  while ( g_running ) {

    g_host = 0;
    if ( hostcount++ > 3 ) {
      g_host = host_events ( 0 );
      hostcount = 0;
    }

    keys = KeyCurrentState();

    if ( g_host & STAT_exeunt ) {
      return ( 0 );
    }

    if ( g_host & STAT_skip ) {
      g_frameskip = host_get_skip ( 0 );
    }

    input.dev[0] = DEVICE_3BUTTON;
    input.pad[0] = 0;

    if ( g_host & STAT_start ) {
      input.pad[0] |= INPUT_START;
    }
    if ( keys & keyBitFunction ) {
      input.pad[0] |= INPUT_START;
    }

    if ( keys & keyBitHard1 || keys & keyBitNavLeft ) {
      input.pad[0] |= INPUT_LEFT;
    } else if ( keys & keyBitHard2 || keys & keyBitNavRight ) {
      input.pad[0] |= INPUT_RIGHT;
    }
    if ( keys & keyBitPageUp ) {
      input.pad[0] |= INPUT_UP;
    } else if ( keys & keyBitPageDown ) {
      input.pad[0] |= INPUT_DOWN;
    }
    if ( keys & keyBitHard3 ) {
      input.pad[0] |= INPUT_A;
    }
    if ( keys & keyBitHard4 ) {
      input.pad[0] |= INPUT_B;
    }
    if ( keys & keyBitActionC || keys & keyBitNavSelect ) {
      input.pad[0] |= INPUT_C;
    }

    if ( frameskip == g_frameskip ) {
      system_frame ( 0 );
      if ( frameskip ) {
	frameskip--;
      }

      // blit if theres a backbuffer in use; if no backbuffer, then its
      // rendering 1:1 right onto the screen itself
      if ( backbuffer ) {

	UInt16 x, y;
	UInt16 xmax = bitmap.viewport.w;
	UInt16 ymax = bitmap.viewport.h;
	UInt16 *fline, *tline;
	UInt16 *f, *t, *tline2, *t2;

	fline = ((UInt16*)bitmap.data) + bitmap.viewport.x;
	tline = ((UInt16*) g_pnobridge -> vram) + ( 32 * display_pitch );
	for ( y = 0; y < ymax; y++ ) {
	  f = fline;
	  t = tline;
	  for ( x = 0; x < xmax; x++ ) {
	    *t++ = *f;
	    if ( x & 1 ) {
	      *t++ = *f;
	    }
	    f++;
	  }
	  fline += bitmap.pitch / 2;
	  tline += display_pitch;
	}

      } else { // backbuffer?
	// no backbuffer, which means its rendering 1:1 to the vram

	// in zodiac case, we're actually rendering 1:1 to the surface, so
	// unlock, blit, and re-lock it
	if ( g_pnobridge -> flags & BF_FULLSCREEN &&
	     g_pnobridge -> flags & BF_ZODIAC )
	{
	  // unlock surface
	  stubzodiacblit ( 2 ); // unlock surface
#if 0
	  // blit surface
	  stubzodiacblit ( 0 ); // blit surface
#else
	  // blit surface with sizing information
	  UInt32 s = bitmap.viewport.w;
	  s |= bitmap.viewport.h << 16;
	  stubzodiacblit ( s ); // blit surface
#endif
	  // pick up new surface
	  bitmap.data = (unsigned char *) (UInt16*) stubzodiacblit ( 1 );
	  bitmap.data -= 64;
	}

      }

    } else {
      system_frame ( 1 );
      if ( frameskip ) {
	frameskip--;
      }
      if ( ! frameskip ) {
	frameskip = g_frameskip;
      }
    }

#ifdef F_FPS
    if ( ! nofps ) {
      // fps counter
      if ( TimGetSeconds() == fps_lasttime ) {
	fps_counter++;
      } else {
	StrIToA ( fps_buffer + 4, fps_counter * 2 );
	fps_counter = 0;
	fps_lasttime = TimGetSeconds();
      }
      WinDrawChars ( "          ", 10, 120, 140 );
      WinDrawChars ( fps_buffer, StrLen ( fps_buffer ), 120, 140 );
    }
#endif

#ifndef AUDIO_ARM
    if ( g_pnobridge -> flags & BF_AUDIO ) {
      audio_gen();
    }
#endif

  } // while running

  /* kill coleco off
   */
  if ( backbuffer ) {
    MemPtrFree ( backbuffer );
  }
  //trash_machine();
  system_shutdown();
  //error_shutdown();
  //MemPtrFree ( cart_rom );
  MemCartFree();

  //DEBUGS ( "Please pinhole reset to clear up memory" );

  // done
  return ( 0 );
}

UInt32 host_events ( UInt32 foo ) {
  UInt32 r;
  UInt32 arg;

  arg = BRIDGE_events | ( foo << BRIDGESHIFT );
  arg = ByteSwap32 ( arg );

  r = g_my_call68KFuncP ( g_my_emulStateP,
			  (UInt32) g_pnobridge -> eventf, &arg, 4 );

  return ( r );
}

UInt32 host_get_skip ( UInt8 foo ) {
  UInt32 r;
  UInt32 arg;

  arg = BRIDGE_get_skip | ( foo << BRIDGESHIFT );
  arg = ByteSwap32 ( arg );

  r = g_my_call68KFuncP ( g_my_emulStateP,
			  (UInt32) g_pnobridge -> eventf, &arg, 4 );

  return ( r );
}

long int StrAToI ( const char *ptr ) {
  UInt8 params [ 4 ];

  params [ 0 ] = (UInt8)((UInt32)(ptr) >> 24);
  params [ 1 ] = (UInt8)((UInt32)(ptr) >> 16);
  params [ 2 ] = (UInt8)((UInt32)(ptr) >> 8);
  params [ 3 ] = (UInt8)(ptr);

  return ((long int)((g_my_call68KFuncP)( (void *)(g_my_emulStateP),
				       (0xA0CE & (0x00000FFF)),
				       &params, sizeof(params))));
}

UInt32 TimGetTicks ( void ) {
  return (( UInt32)((g_my_call68KFuncP)( (void*)(g_my_emulStateP),
					 (0xA0F7 & (0x00000FFF)), 0, 0)));
}

UInt16 SysTicksPerSecond ( void ) {
  return (( UInt16)((g_my_call68KFuncP)( (void*)(g_my_emulStateP),
					 (0xA2E9 & (0x00000FFF)), 0, 0)));
}

UInt32 KeyCurrentState ( void ) {
  return (( UInt32)((g_my_call68KFuncP)( (void*)(g_my_emulStateP),
					 (0xA2A2 & (0x00000FFF)), 0, 0)));
}

void *GetObjectPtr ( UInt16 objectID ) {
  FormPtr frmP;
  frmP = FrmGetActiveForm();
  return FrmGetObjectPtr(frmP, FrmGetObjectIndex(frmP, objectID));
}

void *MemGluePtrNewL ( UInt32 foo ) {
  UInt32 r;
  UInt32 arg;

  arg = BRIDGE_MemGluePtrNew | ( foo << BRIDGESHIFT );
  arg = ByteSwap32 ( arg );

  r = g_my_call68KFuncP ( g_my_emulStateP,
			  (UInt32) g_pnobridge -> eventf, &arg, 4 );

  return ( (UInt8*) r );
}

UInt16 palm_vfs_iter ( void ) {
  return ( g_pnobridge -> vfs_vol );
}

UInt16 palm_vfs_count ( void ) {
  return ( g_pnobridge -> vfs_max );
}

/* -----------------------------------------------------------------------
 */
UInt16 palmap [ 16 ];
#if 0
void SetColor ( byte N, byte R, byte G, byte B ) {
  //XPal [ N ] = ( ( (R) << 11 ) | ( (G) << 5 ) | B );
  XPal [ N ] = N;
  R >>= 3;
  G >>= 2;
  B >>= 3;
  palmap [ N ] = ( ( (R) << 11 ) | ( (G) << 5 ) | B );
  return;
}
#endif

// 0 == blit; 1 == lock; 2 == unlock
UInt32 stubzodiacblit ( UInt32 mode ) {
  UInt32 r;
  UInt32 arg;

  arg = BRIDGE_zodiacblit | ( mode << BRIDGESHIFT );
  arg = ByteSwap32 ( arg );

  r = g_my_call68KFuncP ( g_my_emulStateP,
			  (UInt32) g_pnobridge -> eventf, &arg, 4 );

  return ( r );
}

void stubaudiof ( void *s ) {
  UInt32 r;
  UInt32 arg;
  UInt32 foo = (UInt32) s;

  arg = BRIDGE_accum | ( ((foo&0xFFFF0000)>>16) << BRIDGESHIFT );
  arg = ByteSwap32 ( arg );
  r = g_my_call68KFuncP ( g_my_emulStateP,
			  (UInt32) g_pnobridge -> eventf, &arg, 4 );

  arg = BRIDGE_setaudio | ( (foo&0x0000FFFF) << BRIDGESHIFT );
  arg = ByteSwap32 ( arg );
  r = g_my_call68KFuncP ( g_my_emulStateP,
			  (UInt32) g_pnobridge -> eventf, &arg, 4 );

  return;
}

#if 1
__declspec(pace_native_callback)
#endif
Err audio_callback_extended
  ( void *userData, SndStreamRef stream, void *buffer, UInt32* bufsize )
{
  UInt32 size = *bufsize;
  Int16 *bufptr = buffer;
  UInt8 *sndbuf = (void*) ByteSwap32 ( ((UInt32)userData) );

#ifdef AUDIO_ARM
  // local audio stuff
  UInt16 max;
  UInt8 scale;
  UInt16 frames = *bufsize / AUDIO_STEREO / AUDIO_WIDTH;

  //t_snd *snd = (t_snd*) sndbuf;

  if ( snd.buffer_size < frames ) {
    max = snd.buffer_size;
    scale = frames / snd.buffer_size;
  } else {
    max = frames;
    scale = 1;
  }

  { // gen
    UInt16 i = 0, f = 0, j = 0;
    UInt8 si;

    while ( f < frames/*max*/ ) {
      for ( si = 0; si < scale; si++ ) {
	bufptr [ i++ ] = snd.buffer [ 0 ] [ j ];
	bufptr [ i++ ] = snd.buffer [ 1 ] [ j ];
      }
      f++;
      j++;
      if ( j == max ) {
	j = 0;
      }
    } // while

  } // gen
#endif

  return ( errNone );
}

void error ( char *msg, ... ) {
  debuglog ( msg );
  return;
}

void *MemCart ( void ) {
  UInt32 r;
  UInt32 arg;

  arg = BRIDGE_bigalloc;
  arg = ByteSwap32 ( arg );

  r = g_my_call68KFuncP ( g_my_emulStateP,
			  (UInt32) g_pnobridge -> eventf, &arg, 4 );

  return ( (UInt8*) r );
}

void *MemCartFree ( void ) {
  UInt32 r;
  UInt32 arg;

  arg = BRIDGE_bigfree;
  arg = ByteSwap32 ( arg );

  r = g_my_call68KFuncP ( g_my_emulStateP,
			  (UInt32) g_pnobridge -> eventf, &arg, 4 );

  return ( (UInt8*) r );
}

void audio_gen ( void ) {
  Int16 *buf_ptr = (Int16 *) g_pnobridge -> audio_buf;
  UInt8 *source = (UInt8*) buf_ptr;
  UInt16 max;
  UInt8 scale;

  //return;

#if 1
  if ( source [ 0 ] == 0 ) {
    source [ 0 ] = 1;
    source += 2;
    source += AUDIO_BUFSIZE_BYTES;
  } else {
    source [ 0 ] = 0;
    source += 2;
  }
#endif
  buf_ptr = (Int16*) source;

  // genesis

  //MemMove ( buf_ptr, snd.buffer, snd.buffer_size * 2 );

  if ( snd.buffer_size < AUDIO_BUFSIZE_FRAMES ) {
    max = snd.buffer_size;
    scale = AUDIO_BUFSIZE_FRAMES / snd.buffer_size;
  } else {
    max = AUDIO_BUFSIZE_FRAMES;
    scale = 1;
  }

  { // gen
    UInt16 i = 0, f = 0;
    UInt8 si;

#if 0
    while ( f < max ) {
      for ( si = 0; si < scale; si++ ) {
	buf_ptr [ i++ ] = snd.buffer [ 0 ] [ f % snd.buffer_size ];
	buf_ptr [ i++ ] = snd.buffer [ 1 ] [ f % snd.buffer_size ];
      }
      f++;
    } // while
#endif
    while ( f < max ) {
      buf_ptr [ i++ ] = snd.buffer [ 0 ] [ f % snd.buffer_size ];
      buf_ptr [ i++ ] = snd.buffer [ 1 ] [ f % snd.buffer_size ];
      f++;
    } // while

  } // gen

  return;
}
