\
#include <PalmOS.h>
#include <PCENativeCall.h>
#include <PNOLoader.h>
#include <VfsMgr.h>
#include <PalmOSGlue.h>
#include "StarterRsc.h"
#include "host68k.h"
#include "display68k.h"
#include "pnobridge.h"
#include "starter.h"
#include <Incs/Core/System/DLServer.h>
#include "stdfile.h"

#include "MathLib.h"

#include "zodiac.h"

// if defined, use resource based pnos
//#define RSRC 1

UInt32 query_heap_free ( void );
void garbage_collect ( void );
void fullscreen ( UInt8 activate );

char *g_game_rom = NULL;
UInt8 g_running = 0;

typedef struct _dirent_t {
  char filename [ 128 ];
  struct _dirent_t *next;
} dirent_t;

dirent_t *dirlist = NULL;

SndStreamRef g_sndr = 0; // sound ref
//UInt8 sndbuf [ ( AUDIO_BUFSIZE_BYTES * 2 ) /*( AUDIO_SIZE * 2 * 2 ) */ + 2 ];
UInt8 *sndbuf = NULL;
UInt8 g_ftrmem = 1;
UInt8 g_lowmem = 0;

void dirlist_build ( UInt8 );
char *dirlist_lookup ( UInt16 entry );
void dirlist_dump ( void );
UInt16 dirlist_count ( void );

void GamesDrawList ( Int16 itemNum, RectangleType *bounds, Char **itemsText );

static Err RomVersionCompatible ( UInt32 requiredVersion,
				  UInt16 launchFlags )
{
  UInt32 romVersion;

  // See if we're on in minimum required version of the ROM or later.
  FtrGet(sysFtrCreator, sysFtrNumROMVersion, &romVersion);

  if ( romVersion < requiredVersion ) {
    if ((launchFlags & (sysAppLaunchFlagNewGlobals | sysAppLaunchFlagUIApp)) ==
	(sysAppLaunchFlagNewGlobals | sysAppLaunchFlagUIApp))
    {
      FrmAlert (RomIncompatibleAlert);
		
      // Palm OS 1.0 will continuously relaunch this app unless we switch to 
      // another safe one.
      if (romVersion <= kPalmOS10Version) {
	AppLaunchWithCommand(sysFileCDefaultApp, sysAppLaunchCmdNormalLaunch, NULL);
      }
    }
		
    return sysErrRomIncompatible;
  }

  return errNone;
}

void * GetObjectPtr ( UInt16 objectID ) {
  FormPtr frmP;
  frmP = FrmGetActiveForm();
  return FrmGetObjectPtr(frmP, FrmGetObjectIndex(frmP, objectID));
}

static void MainFormInit ( FormPtr frmP ) {
  ListType *list = (ListType*) GetObjectPtr ( PickerList );
  UInt16 num = palm_game_count ( 0, NULL );

  /* update game list
   */
  LstSetSelection ( list, -1 );
  LstSetDrawFunction ( list, GamesDrawList );
  LstSetListChoices ( list, NULL, num );

  /* load directory listing
   */
  if ( ! dirlist ) {
    dirlist_build ( 255 );
  }

  /* hide controls
   */
  {
    UInt16 pitch, width, height;
    hw_guess_screen_dimensions ( &width, &height, &pitch );
    if ( pitch != 480 ) {
      FrmHideObject ( frmP,
		      FrmGetObjectIndex ( frmP, FullscreenCheckbox ) );
    }
  }

  if ( is_clie() ) {
    ControlType *c;
    c = GetObjectPtr ( AudioCheckbox );
    CtlSetValue ( c, 0 );
  }

  FrmHideObject ( frmP, FrmGetObjectIndex ( frmP, StartButton ) );

  FrmHideObject ( frmP, FrmGetObjectIndex ( frmP, FPSList ) );
  FrmHideObject ( frmP, FrmGetObjectIndex ( frmP, FPSPop ) );

  return;
}

static Boolean MainFormDoCommand ( UInt16 command ) {
  Boolean handled = false;
  FormPtr frmP;

  switch ( command ) {

  case MainOptionsAboutStarterApp:
    MenuEraseStatus ( 0) ;	// Clear the menu status from the display.
    frmP = FrmInitForm (AboutForm);
    FrmDoDialog (frmP);		// Display the About Box.
    FrmDeleteForm (frmP);
    handled = true;
    break;

  }
	
  return handled;
}

static Boolean MainFormHandleEvent ( EventPtr eventP ) {
  Boolean handled = false;
  FormPtr frmP;
  FILE *f;

  switch ( eventP -> eType ) {

  case menuEvent:
    return MainFormDoCommand(eventP->data.menu.itemID);

  case keyDownEvent:

    if ( g_running ) {
      handled = true;
      break;
    }

    if ( eventP -> data.keyDown.chr == pageUpChr ) {
      ListType *list = (ListType*) GetObjectPtr ( PickerList );
      LstScrollList ( list, winUp, LstGetVisibleItems ( list ) );
      handled = true; break;
    } else if ( eventP -> data.keyDown.chr == pageDownChr ) {
      ListType *list = (ListType*) GetObjectPtr ( PickerList );
      LstScrollList ( list, winDown, LstGetVisibleItems ( list ) );
      handled = true; break;
    }

    if ( EvtKeydownIsVirtual ( eventP ) ) {

      if ( eventP -> data.keyDown.chr == vchrRockerDown ) {
	ListType *list = (ListType*) GetObjectPtr ( PickerList );
	LstScrollList ( list, winDown, LstGetVisibleItems ( list ) );
      } else if ( eventP -> data.keyDown.chr == vchrRockerUp ) {
	ListType *list = (ListType*) GetObjectPtr ( PickerList );
	LstScrollList ( list, winUp, LstGetVisibleItems ( list ) );
      }

      handled = true;
      break;
    }

    handled = true;
    break;

  case ctlSelectEvent:

    switch ( eventP -> data.ctlSelect.controlID ) {

    case GoButton:
      handled = true;
      run_emu();
      break;

    case ConsolePop:
      handled = true;
      {
	ListType *list = (ListType*) GetObjectPtr ( ConsolesList );
	Int16 s;
	s = LstPopupList ( list );
	if ( s > -1 ) {
	  // dump old list
	  dirlist_dump();
	  // reload new list
	  // sega 1
	  // sms/gg 2
	  // pce 3
	  // vcs 4
	  if ( s == 0 ) { // all
	    dirlist_build ( 255 ); // all
	  } else if ( s == 1 ) { // sega
	    dirlist_build ( 1 ); // sega
	  } else if ( s == 2 ) { // tg16
	    dirlist_build ( 3 ); // tg16
	  } else if ( s == 3 ) { // vcs
	    dirlist_build ( 4 ); // vcs
	  } else { //sms/gg
	    dirlist_build ( 2 ); //sms/gg
	  }
	  // redraw
	  list = (ListType*) GetObjectPtr ( PickerList );
	  LstSetListChoices ( list, NULL, dirlist_count() );
	  LstDrawList ( list );
	}
      }
      break;

    } // switch

    break;

  case frmOpenEvent:
    frmP = FrmGetActiveForm();
    MainFormInit( frmP );
    FrmDrawForm ( frmP);
    handled = true;
    break;
			
  case frmUpdateEvent:
    // To do any custom drawing here, first call FrmDrawForm(), then do your
    // drawing, and then set handled to true.
    break;

  case frmCloseEvent:
    dirlist_dump();
    break;

  default:
    break;
		
  }
	
  return handled;
}

Boolean AppHandleEvent ( EventPtr eventP ) {
  UInt16 formId;
  FormPtr frmP;

  if (eventP->eType == frmLoadEvent) {
    // Load the form resource.
    formId = eventP->data.frmLoad.formID;
    frmP = FrmInitForm(formId);
    FrmSetActiveForm(frmP);

    // Set the event handler for the form.  The handler of the currently
    // active form is called by FrmHandleEvent each time is receives an
    // event.
    switch (formId) {

    case MainForm:
      FrmSetEventHandler(frmP, MainFormHandleEvent);
      break;

    default:
      // ErrFatalDisplay("Invalid Form Load Event");
      break;

    }
    return true;
  }
	
  return false;
}

static void AppEventLoop ( void ) {
  UInt16 error;
  EventType event;

  do {
    EvtGetEvent(&event, evtWaitForever);

    if (! SysHandleEvent(&event))
      if (! MenuHandleEvent(0, &event, &error))
	if (! AppHandleEvent(&event))
	  FrmDispatchEvent(&event);

  } while (event.eType != appStopEvent);

  return;
}

static Err AppStart ( void ) {
  UInt16 voliter;
  FileRef dirRef = 0;
  Err error;

  /* compact memory
   */
  MemHeapCompact ( 0 );

  /* low heap situation?
   */
  if ( query_heap_free() < 7000000 ) {
    g_ftrmem = 1;
  }
  // force it anyway..
  g_ftrmem = 1;

  // low memory?
  if ( query_heap_free() < 4500000 ) {
    g_lowmem = 1;
  }

  /* math lib
   */
#if 0
  Err error;

  error = SysLibFind ( MathLibName, &MathLibRef );
  if ( error ) {
    error = SysLibLoad ( LibType, MathLibCreator, &MathLibRef );
  }

  if ( error ) {
    DEBUGS ( "Can't find mathlib!" );
    return ( error );
  }
  error = MathLibOpen ( MathLibRef, MathLibVersion );
  if ( error ) {
    DEBUGS ( "Can't open mathlib!" );
    return ( error );
  }
#endif

  /* game rom memory
   */
  g_game_rom = MemPtrNew ( 100 );
  StrCopy ( g_game_rom, "soniche1.smd" );

  /* create our directory if we can't open it..
   */
  voliter = palm_vfs_iter();

  error = VFSFileOpen ( voliter, VFSPATH, vfsModeRead, &dirRef );

  if ( error != errNone ) {
    error = VFSDirCreate ( voliter, VFSROOT );
    error = VFSDirCreate ( voliter, VFSPATH );

    error = VFSFileOpen ( voliter, VFSPATH, vfsModeRead, &dirRef );

    if ( error != errNone ) {
      FrmCustomAlert ( MessageAlert,
		       "Zot requires expansion media for its "
		       "data files; put media in first slot.", NULL, NULL );
      return ( -1 );
    }

  }

  VFSFileClose ( dirRef );

  // register some VFS directories
  VFSRegisterDefaultDirectory ( ".smd", expMediaType_Any, VFSPATH );
  VFSRegisterDefaultDirectory ( ".SMD", expMediaType_Any, VFSPATH );

  /* go to 16bpp
   */
  if ( ! hw_set16bpp() ) {
    DEBUGS ( "Couldn't set to 16 bit colour!" );
    return ( 1 );
  }

  /* zodiac?
   */
#if 1
  if ( is_zodiac() ) {

    g_zodiac = 1;

    if ( ! tw_gfx_init() ) {
      DEBUGS ( "Can't fire up Zodiac hardware!" );
      return ( 1 );
    } else {
      //DEBUGS ( "Turned on Zodiac hardware." );
    }

  } // zodiac
#endif

#if 0
  /* this is a dev build.. stop normal people from using it
   */
  {
    char name [ 35 ];
    UInt32 h;

    query_hotsync_name ( name );

    h = hash_string ( name );

    //DEBUGU32 ( hash_string ( "Jeff Mitchell" ) );
    //DEBUGU32 ( hash_string ( "Orca" ) );
    //DEBUGU32 ( hash_string ( "T3" ) );
    //DEBUGU32 ( hash_string ( "O's Zodiac" ) );
    //DEBUGU32 ( hash_string ( "dustylou" ) );

    switch ( h ) {
    case 0x002EBE5C: // jeff // jeff
    case 0x00464F5D: // orca // jeff
    case 0x000046D7: // T3 // jm
    case 0x0EDB6E87: // O's Zodiac // jm
    case 0x04BBF7B1: // dustylou // daneel
      break;
    default:
      DEBUGS ( "This is a developer-only build of ZOT!" );
      return ( -1 );
    }

  }
#endif

  return errNone;
}

static void AppStop ( void ) {
  // Close all the open forms.
  FrmCloseAllForms ();
}

static
UInt32 StarterPalmMain ( UInt16 cmd, MemPtr /*cmdPBP*/, UInt16 launchFlags ) {
  Err error;

  error = RomVersionCompatible (kOurMinVersion, launchFlags);
  if ( error ) return ( error );

  switch (cmd) {

  case sysAppLaunchCmdNormalLaunch:
    error = AppStart();
    if (error) 
      return error;
				
    FrmGotoForm(MainForm);
    AppEventLoop();
    AppStop();
    break;

  default:
    break;

  }
	
  return errNone;
}

UInt32 PilotMain ( UInt16 cmd, MemPtr cmdPBP, UInt16 launchFlags ) {
  return ( StarterPalmMain ( cmd, cmdPBP, launchFlags ) );
}

UInt32 hash_string ( char *name ) {
  UInt32 g = 0;
  UInt32 h = 0;
  char *temp;
   
  for ( temp = name; *temp; temp++ ) {

    h = (h << 4) + (*temp * 13);

    if ( ( g = h & 0xf0000000 ) ) {
      h ^= (g >> 24);
      h ^= g;
    }

  }

  return ( h );
}

int query_hotsync_name ( char *r_buffer ) {
  char name [ dlkMaxUserNameLength + 1 ];
  // char *name  = MemHandleLock ( MemHandleNew ( dlkMaxUserNameLength + 1 ) );

  MemSet ( name, dlkMaxUserNameLength + 1, '\0' );

  DlkGetSyncInfo ( NULL, NULL, NULL, name, NULL, NULL );

  StrCopy ( r_buffer, name );

  return ( 1 );
}

char *strdup ( char *text ) {
  char *new;
  UInt16 len;

  if ( ! text ) {
    return ( NULL );
  }

  len = StrLen ( text );

  new = MemPtrNew ( len + 1 );

  if ( ! new ) {
    return ( NULL );
  }

  MemSet ( new, len + 1, '\0' );

  StrCopy ( new, text );

  return ( new );
}

void dirlist_build ( UInt8 scope ) {
  UInt16 count = palm_game_count ( 0, NULL );
  UInt16 iter;
  char filename [ 128 ];
  dirent_t *p;

  for ( iter = 0; iter < count; iter++ ) {

    palm_game_count ( iter + 1, filename );

    if ( ( scope == 255 ) ||
	 ( scope == query_filetype ( filename ) ) ||
	 ( query_filetype ( filename ) == 0 ) )
    {
      p = MemPtrNew ( sizeof(dirent_t) );
      StrToLower ( p -> filename, filename );
      //StrCopy ( p -> filename, filename );
      p -> next = NULL;

      if ( dirlist ) {
	dirent_t *foo = dirlist;
	dirent_t *last = NULL;

	while ( ( foo ) &&
		( StrCaselessCompare ( p -> filename, foo -> filename ) > 0 ) )
	{
	  last = foo;
	  foo = foo -> next;
	}

	if ( last ) {
	  last -> next = p;
	  p -> next = foo;
	} else {
	  p -> next = dirlist;
	  dirlist = p;
	}

      } else {
	// first
	dirlist = p;
      }

    } // match

  } // for

  return;
}

char *dirlist_lookup ( UInt16 entry ) {
  dirent_t *iter = dirlist;

  while ( entry ) {
    iter = iter -> next;
    entry--;
  }

  return ( iter -> filename );
}

void dirlist_dump ( void ) {
  dirent_t *iter = dirlist;
  dirent_t *next;

  while ( iter ) {
    next = iter -> next;
    MemPtrFree ( iter );
    iter = next;
  }

  dirlist = NULL;

  return;
}

UInt16 dirlist_count ( void ) {
  dirent_t *iter = dirlist;
  UInt16 c = 0;

  while ( iter ) {
    c++;
    iter = iter -> next;
  }

  return ( c );
}

void GamesDrawList ( Int16 itemNum, RectangleType *bounds, Char **itemsText ) {
  char *filename = NULL;
  ListType *list = (ListType*) GetObjectPtr ( PickerList );
  UInt32 crc;

  // get filename
  filename = dirlist_lookup ( itemNum );

  // draw filename
  if ( filename ) {
    WinDrawChars ( filename, StrLen ( filename ),
		   bounds -> topLeft.x, bounds -> topLeft.y );
  }

  return;
}

void run_emu ( void ) {
  UInt32 processor;
  ListType *list = (ListType*) GetObjectPtr ( PickerList );
  UInt32 flags = 0;
  ControlType *c;
  FormPtr frmP = FrmGetActiveForm();
  UInt8 filetype;

  if ( LstGetSelection ( list ) < 0 ) {
    DEBUGS ( "No Cartridge ROM was selected!" );
    return;
  }

  // flag running
  g_running = 1;

  // file type
  filetype = query_filetype ( dirlist_lookup ( LstGetSelection ( list ) ) );

  if ( filetype == 0 ) {
    // shouldn't occur; GUI blocks unknown ROMs
    DEBUGS ( "Invalid ROM type" );
    return;
  }

  //DEBUGU8 ( filetype );
  //DEBUGS ( emu_module [ filetype ] );

  // does emu module exist?
  {
    FILE *f = fopen ( emu_module [ filetype ], "r" );
    if ( ! f ) {
      DEBUGS ( "The emulation module to run this game is not present." );
      return;
    }
    fclose ( f );
  }

  //DEBUGS ( "heap is" );
  //DEBUGU32 ( query_heap_free() );

  // flags
  c = GetObjectPtr ( FullscreenCheckbox );
  if ( CtlGetValue ( c ) ) {
    flags |= BF_FULLSCREEN;
    fullscreen ( 1 );
  }

  c = GetObjectPtr ( AudioCheckbox );
  if ( CtlGetValue ( c ) ) {
    flags |= BF_AUDIO;
  }
  //flags &= ~BF_AUDIO;

  if ( g_lowmem ) {
    flags |= BF_LOWMEM;
  }

  if ( is_zodiac() ) {
    flags |= BF_ZODIAC;
  }

  // clear background
  {
    RectangleType r;
    IndexedColorType c;

    r.topLeft.x = 0;
    r.topLeft.y = 16;
    r.extent.x = 160;
    r.extent.y = 160 - 16;

    c = WinSetBackColor ( 0 );
    WinDrawRectangle ( &r, 0 );
    WinSetBackColor ( c );
  }

  // hide controls
  FrmHideObject ( frmP, FrmGetObjectIndex ( frmP, GoButton ) );
  FrmHideObject ( frmP, FrmGetObjectIndex ( frmP, PickerList ) );

  // show controls
  FrmShowObject ( frmP, FrmGetObjectIndex ( frmP, StartButton ) );

  // show fps controls
  //FrmShowObject ( frmP, FrmGetObjectIndex ( frmP, FPSList ) );
  FrmShowObject ( frmP, FrmGetObjectIndex ( frmP, FPSPop ) );

  // call the PACE Native Object
  FtrGet(sysFtrCreator, sysFtrNumProcessorID, &processor);

  // rearrange memory
  garbage_collect();
  garbage_collect();

  // set up sound buffer
  sndbuf = MemPtrNew ( ( AUDIO_BUFSIZE_BYTES * 2 ) + 2 );

  // fire up pno
  if ( sysFtrNumProcessorIsARM ( processor ) &&
       LstGetSelection ( list ) >= 0 )
  {
    PnoDescriptor pno;
    pnobridge_t *bridge = (pnobridge_t*) MemPtrNew ( sizeof(pnobridge_t) );
    MemHandle mh;
    MemPtr pnoPtr;
    UInt32 result;
    UInt16 pitch, width, height;
    UInt32 pnosize = 0;
    UInt16 iter = 1;
    UInt8 *pnobuf = NULL;
    UInt8 *pnotarget;
    Err error;

    hw_guess_screen_dimensions ( &width, &height, &pitch );

    bridge -> vfs_max = palm_vfs_count();
    bridge -> vfs_vol = palm_vfs_iter();
    bridge -> vram = (UInt32) hw_query_vram();
    bridge -> eventf = (UInt32) host_event_handler;
    bridge -> display_width = width;
    bridge -> display_height = height;
    bridge -> display_pitch = pitch;
    bridge -> game_rom =
      (UInt32) dirlist_lookup ( LstGetSelection ( list ) );
    bridge -> audio_buf = (UInt32) sndbuf;
    bridge -> flags = flags;

    if ( filetype == 1 ) {
      iter = 1;
    } else if ( filetype == 2 ) {
      iter = 0x64;
    } else if ( filetype == 3 ) {
      iter = 0x128;
    }

#ifdef RSRC
    // how large is pnolet?
    while ( ( mh = DmGetResource ( 'ARMC', iter ) ) ) {
      pnosize += MemHandleSize ( mh );
      iter++;
      DmReleaseResource ( mh );
    }

    // get pnolet memory
    if ( g_ftrmem ) {
      error = FtrPtrNew ( CREATORID, FTR_PNO, pnosize, (void**) &pnobuf );

      if ( error || pnobuf == NULL ) {
	DEBUGS ( "Couldn't allocate enough storage memory for engine!" );
	DEBUGU32 ( pnosize );
	return;
      }

    } else {
      pnobuf = (UInt8*) MemGluePtrNew ( pnosize );

      if ( pnobuf == NULL ) {
	DEBUGS ( "Couldn't allocate enough runtime heap memory for engine!" );
	DEBUGU32 ( pnosize );
	return;
      }

    }
#else // use a bin file instead
    {
      FILE *f = fopen ( emu_module [ filetype ], "r" );
      UInt16 c = 0;
      UInt32 left;
      char msg [ 100 ];
      char *b = MemPtrNew ( 40000 );

      if ( ! f ) {
	DEBUGS ( "Couldn't find emulation module!" );
	return;
      }

      fseek ( f, 0, SEEK_END );
      pnosize = ftell ( f );
      fseek ( f, 0, SEEK_SET );
#if 0
      fclose ( f );
      f = fopen ( emu_module [ filetype ], "r" );
      if ( ! f ) {
	DEBUGS ( "Messed up." );
	return;
      }
#endif

      // DEBUGU32 ( pnosize );

      // get pnolet memory
#if 1
      error = FtrPtrNew ( CREATORID, FTR_PNO, pnosize, (void**) &pnobuf );

      if ( error || pnobuf == NULL ) {
	DEBUGS ( "Couldn't allocate enough storage memory for module!" );
	DEBUGU32 ( pnosize );
	return;
      }

      left = pnosize;
      pnotarget = pnobuf;
      while ( left ) {
	UInt16 s;
	if ( left > 40000 ) {
	  s = 40000;
	} else {
	  s = left;
	}

	// log
	StrPrintF ( msg, "Assembling emulator core [ %u ]", c );
	WinDrawChars ( msg, StrLen ( msg ), 0, 150 );
	c++;

	fread ( b, 1, s, f );
	DmWrite ( pnobuf, pnotarget - pnobuf, b, s );

	pnotarget += s;
	left -= s;
      } // while

      MemPtrFree ( b );
#else
      pnobuf = (UInt8*) MemGluePtrNew ( pnosize );

      if ( ! pnobuf ) {
	DEBUGS ( "Couldn't allocate enough runtime memory for engine!" );
	return;
      }

      fread ( pnobuf, 1, pnosize, f );
#endif

      fclose ( f );

    }
#endif

    //StrPrintF ( buf, "Allocated %lu size pnolet.", pnosize );
    //DEBUGS ( buf );

#ifdef RSRC
    if ( filetype == 1 ) {
      iter = 1;
    } else if ( filetype == 2 ) {
      iter = 0x64;
    } else if ( filetype == 3 ) {
      iter = 0x128;
    }

    // reset
    pnotarget = pnobuf;

    // read pnolet into buffer
    while ( ( mh = DmGetResource ( 'ARMC', iter ) ) ) {
      char msg [ 100 ];

      // lock source
      UInt8 *p = (UInt8*) MemHandleLock ( mh );

      // copy data
      if ( g_ftrmem ) {
	DmWrite ( pnobuf, pnotarget - pnobuf, p, MemHandleSize ( mh ) );
      } else {
	MemMove ( pnotarget, p, MemHandleSize ( mh ) );
      }

      // log
      //StrPrintF ( msg, "Assembling emulator core [ %u ]", iter );
      //WinDrawChars ( msg, StrLen ( msg ), 0, 150 );

      // unlock
      MemHandleUnlock ( mh );

      // log
      StrPrintF ( msg, "Assembling emulator core [ %u ]", iter );
      WinDrawChars ( msg, StrLen ( msg ), 0, 150 );

      // next!
      pnotarget += MemHandleSize ( mh );
      iter++;
      DmReleaseResource ( mh );
    }
#endif

    //mh = DmGetResource('ARMC', 1);
    //pnoPtr = MemHandleLock(mh);
    pnoPtr = pnobuf;

    // begin audio
    if ( flags & BF_AUDIO ) {

      if ( filetype == 1 ) {
#ifndef AUDIO_ARM
	audio_begin ( audio_callback_extended, false, NULL );
#endif
      } else if ( filetype == 2 ) {
	audio_begin ( audio_callback_extended, false, NULL );
      } else if ( filetype == 3 ) {
	audio_begin ( audio_callback_extended, false, NULL );
      } else if ( filetype == 4 ) {
	audio_begin ( audio_callback_extended, false, NULL );
      }

    }

    // fire up pnolet
    PnoLoad ( &pno, pnoPtr );
    result = PnoCall ( &pno, bridge );
    PnoUnload ( &pno );

    // end audio
    if ( g_sndr /* flags & BF_AUDIO */ ) {
      audio_end();
    }

    //DEBUGU32 ( result );

    //MemHandleUnlock(mh);
    //DmReleaseResource(mh);
    if ( g_ftrmem ) {
      FtrPtrFree ( CREATORID, FTR_PNO );
    } else {
      MemPtrFree ( pnobuf );
    }

    // big genesis cart memory left behind?
    {
      extern UInt32 g_bigalloc;
      if ( g_bigalloc ) {
	MemPtrFree ( (void*)g_bigalloc );
      }
    }

    // rearrange memory
    garbage_collect();

  } // is arm and rom picked

  return;
}

void audio_begin ( void *f, UInt8 arm, void *d ) {
  Err error;

  if ( g_sndr ) {
    return; // already doing audio!
  }

  error = SndStreamCreateExtended
    ( &g_sndr, sndOutput, sndFormatPCM, AUDIO_RATE, /*sndUInt8*/ AUDIO_FMT,
      AUDIO_STEREO == 2 ? sndStereo : sndMono,
      f, d ? d : sndbuf, AUDIO_BUFSIZE_FRAMES, arm /*true*/ /*false*/ );

  if ( error ) {
    DEBUGS ( "Couldn't open audio stream!" );
    DEBUGU32(error);
  } else {
    MemSet ( sndbuf, ( AUDIO_BUFSIZE_BYTES * 2 ) + 2 , '\0' );
    SndStreamSetVolume ( g_sndr, sndGameVolume /* 25 * 64 */ );
    //audio_volume ( 128 );
    SndStreamStart ( g_sndr );
  }

  return;
}

void audio_end ( void ) {
  if ( g_sndr ) {
    SndStreamStop ( g_sndr );
    SndStreamDelete ( g_sndr );
  }
  return;
}

/* SDK5.2 offers Extended audio callbacks; much more useful than the basic
 * audio system...
 */
Err audio_callback_extended
  ( void *userData, SndStreamRef stream, void *buffer, UInt32* bufsize )
{
  UInt8 *source = userData;
  //UInt8 set = source [ ( AUDIO_SIZE * 2 * 2 ) ];
  UInt8 set = source [ 0 ];

#if 1 // alternate between two buffers
  /* check first first byte to find out which buffer is the one to read; the
   * other one is being written to right now
   */

  // audio ringbuffer
  if ( set == 1 ) {
    // working on 1, do set 0
    source += 2;
  } else {
    // working on 0, do set 1
    source += 2;
    source += AUDIO_BUFSIZE_BYTES;
  }

  if ( *bufsize <= AUDIO_BUFSIZE_BYTES ) {
    MemMove ( buffer, source, *bufsize );
  } else {
    MemMove ( buffer, source, AUDIO_BUFSIZE_BYTES );
    *bufsize = AUDIO_BUFSIZE_BYTES;
  }
#endif


#if 0 // WORKS WELL alternate between two buffers
  /* check first first byte to find out which buffer is the one to read; the
   * other one is being written to right now
   */

  // audio ringbuffer
  if ( set == 1 ) {
    // working on 1, do set 0
    source += 2;
  } else {
    // working on 0, do set 1
    source += 2;
    source += ( AUDIO_SIZE * 2 );
  }

  // working well
#if 0
  if ( *bufsize <= AUDIO_SIZE ) {
    MemMove ( buffer, source, *bufsize * 2 );
  } else {
    MemMove ( buffer, source, AUDIO_SIZE * 2 );
    *bufsize = AUDIO_SIZE;
  }
#endif
  if ( *bufsize <= (AUDIO_SIZE*2) ) {
    MemMove ( buffer, source, *bufsize );
  } else {
    MemMove ( buffer, source, AUDIO_SIZE * 2 );
    *bufsize = AUDIO_SIZE * 2;
  }
#endif

#if 0 // set other-buffer for writer
  /* check first first byte to find out which buffer is the one to read; the
   * other one is being written to right now
   */

  // set writer to use other-buffer
  source [ 0 ] = ( set + 1 ) & 1;

  // audio ringbuffer
  source += 2;
  if ( set == 1 ) {
    source += ( AUDIO_SIZE * 2 );
  }

  // working well
  if ( *bufsize <= AUDIO_SIZE ) {
    MemMove ( buffer, source, *bufsize * 2 );
  } else {
    MemMove ( buffer, source, AUDIO_SIZE * 2 );
    *bufsize = AUDIO_SIZE;
  }
#endif

#if 0 // fill from specified buffer
  /* check first first byte to find out which buffer is the one to read; the
   * other one is being written to right now
   */

  // audio ringbuffer
  if ( set == 0 ) {
    // working on 1, do set 0
    source += 2;
  } else {
    // working on 0, do set 1
    source += 2;
    source += ( AUDIO_SIZE * 2 );
  }

  // working well
  if ( *bufsize <= AUDIO_SIZE ) {
    MemMove ( buffer, source, *bufsize * 2 );
  } else {
    MemMove ( buffer, source, AUDIO_SIZE * 2 );
    *bufsize = AUDIO_SIZE;
  }
#endif

#if 0 // fill from first finished buffer
  /* check first first byte to find out which buffer is the one to read; the
   * other one is being written to right now
   */

  // audio ringbuffer
  if ( set & 1 ) {
    // buffer 1 is done, so use it
    source += 2;
  } else if ( set & 2 ) {
    // buffer 2 is done, so use it
    source += 2;
    source += ( AUDIO_SIZE * 2 );
  } else {
    // buffer underrun
    if ( source [ 1 ] > 250 ) {
      //DEBUGS ( "underrun" );
      source [ 1 ] = 0;
    }
    source [ 1 ] ++;
    MemSet ( buffer, AUDIO_SIZE, '\0' );
    return ( false );
  }

  // working well
  //if ( *bufsize <= AUDIO_SIZE ) {
  //MemMove ( buffer, source, *bufsize * 2 );
  //} else {
    MemMove ( buffer, source, AUDIO_SIZE * 2 );
    *bufsize = AUDIO_SIZE * 2;
    //}

  source = userData;
  if ( set & 1 ) {
    // working on 1, do set 0
    set &= ~1;
  } else if ( set & 2 ) {
    // working on 0, do set 1
    set &= ~2;
  }
  source [ 0 ] = set;
#endif

  return ( false );
}

UInt32 query_heap_free ( void ) {
  UInt16 heapid = 0;
  UInt32 max, bytes;

  MemHeapFreeBytes ( heapid, &bytes, &max );

  //DEBUGS ( "heapfree" );
  //DEBUGU32 ( bytes );

  return ( max );
}

/* garbage_collect() will try and free up memory until the low memory
 * threshhold has been freed. If it cannot get memory above this point,
 * a warning is issued.
 */
void garbage_collect ( void ) {

  /* on failure, compact memory */
  //MemHeapCompact ( MemPtrHeapID ( true_head() ) );

#ifdef RSRC
  MemHeapCompact ( 0 );
#endif

  return;
}

void fullscreen ( UInt8 activate ) {
  UInt8 os5highres = 0;
  UInt32 winvers = 0;
  FormPtr frmP = FrmGetActiveForm();

  if ( ( FtrGet ( sysFtrCreator, sysFtrNumWinVersion, &winvers ) == 0 ) &&
       ( winvers >= 4 ) )
  {
    /* high density api is available! Check if density is present */
    UInt16 density = 0;

    /* high density feature set available; is high resolution available? */
    WinGetSupportedDensity ( &density );

    if ( density == kDensityDouble ) {
      os5highres = 1;
    } else {

      while ( density != 0 ) {

	WinGetSupportedDensity ( &density );

	if ( density == kDensityDouble ) {
	  os5highres = 1;
	  break;
	}

      } /* while */

    } /* density */

    if ( os5highres ) {
      UInt32 version;
      Err err = FtrGet ( pinCreator, pinFtrAPIVersion, &version );
      if ( ! err && version ) {

	// SDK 5.3 says to use Custom..
	//FrmSetDIAPolicyAttr ( frmP, frmDIAPolicyStayOpen );
	FrmSetDIAPolicyAttr ( frmP, frmDIAPolicyCustom );

	// Enable the input trigger
	PINSetInputTriggerState ( pinInputTriggerEnabled );

	// set up constraints
	WinSetConstraintsSize ( FrmGetWindowHandle ( frmP ),
				160, 160, 225, 160, 160, 225 );

	// are we returning to form with a cached fullscreen flag?
	PINSetInputAreaState ( pinInputAreaClosed );

	// hide status bar
	StatHide();

      }
    }

  }

  return;
}
