
#include "palmdep.h"
#include "endianutils.h"
#include "pnobridge.h"
#include "pcx.h"

#include "oscompat.h"

int errno = 0;

#define F_DEBUGLOG 1
void debuglog ( char *poo ) {
#ifdef F_DEBUGLOG
#define MAXDLOG 3
  static char debug [ MAXDLOG ][ 100 ];
  static UInt16 debuglogn = 0;
  static UInt8 nomore = 0;
  char *c;

  if ( ! poo ) {
    nomore = 1;
  }

  if ( nomore ) {
    return; // stopped
  }

  // quick draw to top
  //WinDrawChars ( poo, StrLen ( poo ), 0, 0 );
  //return;

  // normal debuglog
  if ( c = StrChr ( poo, '\n' ) ) {
    *c = '\0';
  }

#if 1
  Int8 iter;
  UInt8 y = 0;

  // dumb
  if ( debuglogn == MAXDLOG - 1 ) {
    for ( iter = 1; iter < MAXDLOG; iter++ ) {
      StrNCopy ( debug [ iter - 1 ], debug [ iter ], 99 );
    }
  }

  MemSet ( debug [ debuglogn ], 100, '\0' );
  StrNCopy ( debug [ debuglogn ], poo, 97 );
  StrNCat ( debug [ debuglogn ], "   ", 99 );

  if ( debuglogn < MAXDLOG - 1 ) {
    debuglogn++;
  }

  // display
  for ( iter = 0; iter < MAXDLOG; iter++ ) {
    char *text = debug [ iter ];
    if ( text ) {
      WinDrawChars ( "                              "
		     "                              "
		     "                              ",
		     90, 0, ( 10 * iter ) + 130 );
      WinDrawChars ( text, StrLen ( text ), 0, ( 10 * iter ) + ( 160 - ( 10 * MAXDLOG ) ) );
    }
  }

#if 0
  {
    UInt32 timmy = TimGetSeconds();
    while ( TimGetSeconds() == timmy );
  }
#endif

#else
  FrmCustomAlert ( MessageAlert, poo, NULL, NULL );
#endif

#endif
  return;
}

void *memcpy ( void *d, const void *s, unsigned int size ) {
  MemMove ( d, s, size );
  return ( d );
}

void *memset ( void *b, int c, unsigned int len ) {
  MemSet ( b, len, c );
  return ( b );
}

void *malloc ( unsigned int s ) {
  //DEBUGU8 ( s );
  void *p = MemGluePtrNewL ( s );
  if ( ! p ) {
    DEBUGS ( "Couldn't allocate memory!" );
    //DEBUGU32 ( s );
  }
  return ( p );
}

void *calloc ( unsigned int s, unsigned int n ) {
  UInt8 *p = MemGluePtrNewL ( s * n );
  memset ( p, 0, s * n );
  return ( p );
}

void free ( void *p ) {
  MemPtrFree ( p );
}

void abort ( void ) {
  DEBUGS ( "PNG abort" );
}

int abs ( int x ) {
  if ( x < 0 ) {
    return ( x * -1 );
  }
  return ( x );
}

void *realloc ( UInt8 *p, UInt32 s ) {
  UInt8 *newbie;

  newbie = MemGluePtrNewL ( s );

  if ( ! newbie ) {
    DEBUGS ( "realloc crash" );
    return ( NULL );
  }

  if ( ! p ) {
    return ( newbie );
  }

  if ( s < MemPtrSize ( p ) ) {
    MemMove ( newbie, p, s );
  } else {
    MemMove ( newbie, p, MemPtrSize ( p ) );
  }

  MemPtrFree ( p );
  //MemPtrResize ( p, s );
  return ( newbie );
}

void raise ( int ) {
  DEBUGS ( "raise error" );
  while(1);
  return;
}

static float math_sine100_deg [ 360 ] = {
  0, 1, 3, 5, 6, 8, 10, 12, 
  13, 15, 17, 19, 20, 22, 24, 25, 27, 
  29, 30, 32, 34, 35, 37, 39, 40, 42, 
  43, 45, 46, 48, 49, 51, 52, 54, 55, 
  57, 58, 60, 61, 62, 64, 65, 66, 68, 
  69, 70, 71, 73, 74, 75, 76, 77, 78, 
  79, 80, 81, 82, 83, 84, 85, 86, 87, 
  88, 89, 89, 90, 91, 92, 92, 93, 93, 
  94, 95, 95, 96, 96, 97, 97, 97, 98, 
  98, 98, 99, 99, 99, 99, 99, 99, 99, 
  99, 99, 99, 99, 99, 99, 99, 99, 99, 
  99, 98, 98, 98, 97, 97, 97, 96, 96, 
  95, 95, 94, 94, 93, 92, 92, 91, 90, 
  89, 89, 88, 87, 86, 85, 84, 83, 82, 
  81, 80, 79, 78, 77, 76, 75, 74, 73, 
  72, 70, 69, 68, 67, 65, 64, 63, 61, 
  60, 58, 57, 56, 54, 53, 51, 50, 48, 
  47, 45, 43, 42, 40, 39, 37, 35, 34, 
  32, 31, 29, 27, 26, 24, 22, 20, 19, 
  17, 15, 14, 12, 10, 8, 7, 5, 3, 
  1, 0, -1, -3, -5, -6, -8, -10, -12, 
  -13, -15, -17, -18, -20, -22, -24, -25, -27, 
  -29, -30, -32, -34, -35, -37, -38, -40, -42, 
  -43, -45, -46, -48, -49, -51, -52, -54, -55, 
  -57, -58, -60, -61, -62, -64, -65, -66, -68, 
  -69, -70, -71, -72, -74, -75, -76, -77, -78, 
  -79, -80, -81, -82, -83, -84, -85, -86, -87, 
  -88, -89, -89, -90, -91, -91, -92, -93, -93, 
  -94, -95, -95, -96, -96, -96, -97, -97, -98, 
  -98, -98, -98, -99, -99, -99, -99, -99, -99, 
  -99, -99, -99, -99, -99, -99, -99, -99, -99, 
  -99, -98, -98, -98, -97, -97, -97, -96, -96, 
  -95, -95, -94, -94, -93, -92, -92, -91, -90, 
  -89, -89, -88, -87, -86, -85, -84, -84, -83, 
  -82, -81, -80, -78, -77, -76, -75, -74, -73, 
  -72, -70, -69, -68, -67, -65, -64, -63, -61, 
  -60, -59, -57, -56, -54, -53, -51, -50, -48, 
  -47, -45, -44, -42, -40, -39, -37, -36, -34, 
  -32, -31, -29, -27, -26, -24, -22, -21, -19, 
  -17, -15, -14, -12, -10, -9, -7, -5, -3, 
  -2
};

#if 1
double log10 ( double x ) {
  //return ( log10f ( x ) );
  return ( x );
}

double sin ( double x ) {
  //return ( sinf ( x ) );
  //  return ( x / 2 );
  return ( math_sine100_deg [ (UInt16)x % 360 ] / 100 );
}

double pow ( double x, double y ) {
  //return ( powf ( x, y ) );
#if 1
  UInt16 e = y;

  while ( e-- ) {
    x *= x;
  }

  return ( x );
#endif
}

double log ( double x ) {
  //return ( log10f ( x ) );
  return ( x );
}
#endif
