/*
    Copyright (C) 1998, 1999, 2000, 2001, 2002, 2003  Charles Mac Donald

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "palmdep.h"

#include "shared.h"

#include "cycloneoff.h"

UInt32 cart_rom_size = 0x400000;   // to allow for smaller roms

//uint8 cart_rom[0x400000];   /* Cartridge ROM */
uint8 *cart_rom;   /* Cartridge ROM */
uint8 work_ram[0x10000];    /* 68K work RAM */
uint8 zram[0x2000];         /* Z80 work RAM */
uint8 zbusreq;              /* /BUSREQ from Z80 */
uint8 zreset;               /* /RESET to Z80 */
uint8 zbusack;              /* /BUSACK to Z80 */
uint8 zirq;                 /* /IRQ to Z80 */
uint32 zbank;               /* Address of Z80 bank window */
uint8 gen_running;
uint8 emul_z80=0;


#ifdef __CYCLONE__
struct Cyclone CycloneCPU;
int32 total_cycles;
#endif

unsigned char *saveram = NULL; // The actual saveram buffer
unsigned save_start = 0, save_len = 0; // Start address and length
int save_prot = 0, save_active = 0; // Flags set from $A130F1
/*--------------------------------------------------------------------------*/
/* Init, reset, shutdown functions                                          */
/*--------------------------------------------------------------------------*/

#ifdef __CYCLONE__
/**************************************************/
/********* CYCLONE STUFF **************************/
/**************************************************/

int CycloneDoRun(int cyc)
{

  //DEBUGS ( "dorun");

  if (cyc<=0) return cyc;
  CycloneCPU.cycles=cyc;
  CycloneRun(&CycloneCPU);
  
  return cyc-CycloneCPU.cycles;
}

int CycloneInterrupt(int irq)
{
  //DEBUGS ( "int");
/*  if (reg[1]&0x20)*/ CycloneCPU.irq=(unsigned char)irq;
  return 0;
}

int CycloneMemBase(uint32 pc)
{
  int membase=0;

  //DEBUGS ( "membase");

  if (pc<cart_rom_size)
  {
    membase=(int)cart_rom; // Program Counter in Rom
  }
  else if ((pc&0xe00000)==0xe00000)
  {
    membase=(int)work_ram-(pc&0xff0000); // Program Counter in Ram
  }
  else
  {
    // Error - Program Counter is invalid
    membase=(int)cart_rom;
  }

  return membase;
}
uint32 CycloneCheckPc(uint32 pc)
{

  //DEBUGS ( "check pc");

  pc-=CycloneCPU.membase; // Get real pc
  pc&=0xffffff;

  CycloneCPU.membase=CycloneMemBase(pc);

  return CycloneCPU.membase+pc;
}

// -----------------------------------------------------------------
//                     Read Rom and read Ram


uint8 CycloneRead8(uint32 a)
{

  //DEBUGS ( "read 8");

  a&=0xffffff;
//  if (a<rom_size) return *(uint8 *)(cart_rom+(a^1)); // Rom
//  if ((a&0xe00000)==0xe00000) return *(uint8 *)(work_ram+((a^1)&0xffff)); // Ram

 if (a<0xa00000) {
    // Saveram
    if((save_active)     && (save_len) &&
       (a >= save_start) && ((a - save_start) < save_len)) {
      return saveram[(a^1) - save_start];
    }
    else if ((signed)(a^1)<cart_rom_size) return *(uint8 *)(cart_rom+(a^1));
  }

  return (uint8)m68k_read_memory_8(a);
}

uint16 CycloneRead16(uint32 a)
{

  //DEBUGS ( "read 16");

  a&=0xfffffe;
//  if (a<rom_size) { uint16 *pm=(uint16 *)(cart_rom+a); return pm[0]; } // Rom
//  if ((a&0xe00000)==0xe00000) { uint16 *pm=(uint16 *)(work_ram+(a&0xffff)); return pm[0]; } // Ram

  return m68k_read_memory_16(a);
}

uint32 CycloneRead32(uint32 a)
{

  //DEBUGS ( "read 32");

  a&=0xfffffe;

//  if (a<rom_size) { uint16 *pm=(uint16 *)(cart_rom+a); return (pm[0]<<16)|pm[1]; } // Rom
//  if ((a&0xe00000)==0xe00000) { uint16 *pm=(uint16 *)(work_ram+(a&0xffff)); return (pm[0]<<16)|pm[1]; } // Ram
  return m68k_read_memory_32(a);
}

// -----------------------------------------------------------------
//                            Write Ram

void CycloneWrite8(uint32 a,uint8 d)
{

  //DEBUGS ( "write 8");

//  if ((a&0xe00000)==0xe00000) { uint8 *pm=(uint8 *)(work_ram+((a^1)&0xffff)); pm[0]=d; return; } // Ram
  a&=0xffffff;
  
  // Saveram write
  if(save_len && !save_prot) if(a >= save_start)
    if ((a - save_start) < save_len)
      { saveram[(a^1) - save_start] = d; return; }

  m68k_write_memory_8(a,d);
}

void CycloneWrite16(uint32 a,uint16 d)
{
  //DEBUGS ( "write 16");

//  if ((a&0xe00000)==0xe00000) { *(uint16 *)(work_ram+(a&0xfffe))=d; return; } // Ram
  a&=0xfffffe;
  m68k_write_memory_16(a,d);
}

void CycloneWrite32(uint32 a,uint32 d)
{

  //DEBUGS ( "write 32");

/*  if ((a&0xe00000)==0xe00000)
  {
    // Ram:
    uint16 *pm=(uint16 *)(work_ram+(a&0xfffe));
    pm[0]=(uint16)(d>>16); pm[1]=(uint16)d;
    return;
  }
*/
  a&=0xfffffe;
  m68k_write_memory_32(a,d);
}

#endif
/**************************************************/
/**************************************************/
/**************************************************/
unsigned int vdp_int_ack_callback(unsigned int int_level);

void gen_init(void)
{
    sound_init();
    memset(&snd, 0, sizeof(snd));

    bswap(cart_rom, cart_rom_size/*0x400000*/);
    
    
#ifdef __CYCLONE__    
    memset(&CycloneCPU,0,sizeof(CycloneCPU));                   
	CycloneCPU.checkpc=CycloneCheckPc;
	CycloneCPU.fetch8 =CycloneCPU.read8 =CycloneRead8;
	CycloneCPU.fetch16=CycloneCPU.read16=CycloneRead16;
	CycloneCPU.fetch32=CycloneCPU.read32=CycloneRead32;
	CycloneCPU.write8 =CycloneWrite8;
	CycloneCPU.write16=CycloneWrite16;
	CycloneCPU.write32=CycloneWrite32;
	CycloneCPU.irq_callback=vdp_int_ack_callback;

	__asm 
    {
    	LDR 	R0, = CycloneCPU
  	    ADD		R0,R0,R10
  		STR	 	R9,[R0,#0x8C]
  		STR	 	R10,[R0,#0x90]
    } 

#else
	m68k_set_cpu_type(M68K_CPU_TYPE_68000);
    m68k_pulse_reset();    
#endif

//    error("PC:%08X\tSP:%08X\n", m68k_get_reg(NULL,M68K_REG_PC), m68k_get_reg(NULL,M68K_REG_SP));
    gen_running = 1;
}

void gen_reset(void)
{
    /* Clear RAM */
    memset(work_ram, 0, sizeof(work_ram));
    memset(zram, 0, sizeof(zram));

    gen_running = 1;
    zreset  = 0;    /* Z80 is reset */
    zbusreq = 0;    /* Z80 has control of the Z bus */
    zbusack = 1;    /* Z80 is busy using the Z bus */
    zbank   = 0;    /* Assume default bank is 000000-007FFF */
    zirq    = 0;    /* No interrupts occuring */

    io_reset();

    /* Reset the 68000 emulator */
    
    
#ifdef __CYCLONE__
    CycloneCPU.srh =0x27; // Supervisor mode
    CycloneCPU.a[7]=CycloneCPU.read32(0); // Stack Pointer
    CycloneCPU.membase=0;
    CycloneCPU.pc=CycloneCPU.checkpc(CycloneCPU.read32(4)); // Program Counter
    total_cycles=0;
#else
	m68k_pulse_reset();
#endif	
    //error("PC:%08X\tSP:%08X\n", m68k_get_reg(NULL,M68K_REG_PC), m68k_get_reg(NULL,M68K_REG_SP));    
    

    z80_reset(0);
    z80_set_irq_callback(z80_irq_callback);

}

void gen_shutdown(void)
{
    int i;

    /* Print 68K and Z80 state */
    error("\n");
//    error("\nPC:%08X\tSP:%08X\tSR:%04X\n", m68k_get_reg(NULL,M68K_REG_PC), m68k_get_reg(NULL,M68K_REG_SP), m68k_get_reg(NULL, M68K_REG_SR));
    for(i=0;i<8;i++)
    {
//        error("D%d:%08X\tA%d:%08X\n", 
//            i, m68k_get_reg(NULL, M68K_REG_D0+i),
//            i, m68k_get_reg(NULL, M68K_REG_A0+i));
    }

    error("\n");
    error("PC:%04X\tSP:%04X\n", z80_get_reg(Z80_PC), z80_get_reg(Z80_SP));
    error("AF:%04X\tAF:%04X\n", z80_get_reg(Z80_AF), z80_get_reg(Z80_AF2));
    error("BC:%04X\tBC:%04X\n", z80_get_reg(Z80_BC), z80_get_reg(Z80_BC2));
    error("DE:%04X\tDE:%04X\n", z80_get_reg(Z80_DE), z80_get_reg(Z80_DE2));
    error("HL:%04X\tHL:%04X\n", z80_get_reg(Z80_HL), z80_get_reg(Z80_HL2));
    error("IX:%04X\tIY:%04X\n", z80_get_reg(Z80_IX), z80_get_reg(Z80_IY));
    error("\n");
}

/*--------------------------------------------------------------------------*/
/* Bus controller chip functions                                            */
/*--------------------------------------------------------------------------*/

int gen_busack_r(void)
{
    return (zbusack & 1);
}

void gen_busreq_w(int state)
{
    zbusreq = (state & 1);
    zbusack = 1 ^ (zbusreq & zreset);

    if(emul_z80&&(zbusreq == 0 && zreset == 1))
    {
//#ifdef __EMUL_Z80__    
        z80_execute(32);
//#endif        
    }
}

void gen_reset_w(int state)
{
    zreset = (state & 1);
    zbusack = 1 ^ (zbusreq & zreset);


    if(zreset == 0)
    {
        if(snd.enabled)
        {
            YM2612ResetChip(0);
        }
//#ifdef __EMUL_Z80__
        z80_reset(0);
        z80_set_irq_callback(z80_irq_callback);
//#endif            
    }

}


void gen_bank_w(int state)
{
    zbank = ((zbank >> 1) | ((state & 1) << 23)) & 0xFF8000;
}



int z80_irq_callback(int param)
{
    zirq = 0;
    z80_set_irq_line(0, CLEAR_LINE);
    return 0xFF;
}

unsigned int vdp_int_ack_callback(unsigned int int_level)
{
    switch(int_level)
    {
        case 4: 
            hint_pending = 0;
            break;

        case 6: 
            status &= ~0x0080;
            vint_pending = 0;
            break;
    }
#ifdef __CYCLONE__
    return 0;//M68K_INT_ACK_AUTOVECTOR;
#else
	return M68K_INT_ACK_AUTOVECTOR;
#endif    
}


void bswap(uint8 *mem, int length)
{
    int i;
    for(i = 0; i < length; i += 2)
    {
        uint8 temp = mem[i+0];
        mem[i+0] = mem[i+1];
        mem[i+1] = temp;
    }
}

