/*

- MALLOC!
- Fix collision
* Fix sprites alignment
* Settings
* Files (and backswitching) with rom DB
* Keys config (B as shift for joystick 2)
* Make difficulty switches switchable with L/R
* Check if rom is loaded
* Stop sound on menu
* FPS (using sprintf)
* Menu
* Make controls more responsive
* Frameskip
* Make writes to vscreen go directly to screen
* Make writes to screen directly rotated
* Keys: remove keys[256]
* Lo schermo non va updatato dopo ogni istruzione ma solo quando viene deciso (vblank?)
* Palette

- NOTE: vscreen is only used in DISPLAY.C and RASTER.C
*/

#define RELEASE

#include <stdlib.h>
#include <stdio.h>		/* Just for sprintf */
#include "gpdef.h"
#include "gpstdlib.h"
#include "gpgraphic.h"
#include "vcs32.h"

#include "types.h"
#include "files.h"
#include "vmachine.h"
#include "display.h"
#include "keyboard.h"
#include "config.h"
#include "string.h"
#include "extern.h"

/*our background and sprites generated from a bmp by gp32 converter*/
/*#include "background.h"*/
/*#include "sprites.h"*/
#include "..\\gen\\types.h"

#include "vcs322.h"
#include "gpfilereq.h"
#include "tiasound.h"
#include "gpmm.h"
#include "gpsoundbuf.h"

#include "menubg.h"
#include "menubg_pal.h"

/*our palette*/
GP_HPALETTE h_pal, h_pal_menu;
GP_PALETTEENTRY palette_menu[256];

#define STICK 	0x01
#define PADDLE 	0x02
#define KEYPAD  0x03

#define MENUOFFSET_X	5
#define MENUOFFSET_Y	20
#define MENUKERNING	14

#define MENUTEXT_COL_NORMAL	0
#define MENUTEXT_COL_HEADER	220
#define MENUTEXT_COL_SELECT	10

/* Menu ID definitions. Must always be >0 */
#define MENU_EMU	0	/* get back to the emu */
#define MENU_MAIN	1
#define MENU_FILE	2
#define MENU_SETTINGS	3
#define MENU_ABOUT	4
#define MENU_HELP	5
#define MENU_SETTINGSFRAMESKIP	6
#define MENU_SETTINGSBRIGHTNESS	7
#define MENU_SETTINGSFRAMESKIP0	8
#define MENU_SETTINGSFRAMESKIP1	9
#define MENU_SETTINGSFRAMESKIP2	10
#define MENU_SETTINGSFRAMESKIP3	11
#define MENU_SETTINGSFRAMESKIP4	12

#define MENU_MOVEDELAY	100
#define MENU_MOVEDELAY_FAST	5
#define MENU_SCREENSIZE	15

typedef struct FINFO
  {
    char name[42];
    int size;
    BOOL isdir;
  } FINFO;

FINFO ** pdirlist;
ulong num_entries;

unsigned short nflip;
int ExKey;
extern keys[256];

char FPS_string[64];
unsigned char current_screen;
unsigned long current_menu;
short menu_cursor=0;
#ifdef RELEASE
char VCSbasedir[] = "gp:\\gpmm\\vcs";
#else
char VCSbasedir[] = "gp:\\\\gpmm";
#endif
char menu_filename[12];

char cart_loaded = 0;

unsigned short frames_count;

FSEL filesel;
char filebuffer[256];
/*PCM_SR pcmsound;*/
GPSOUNDBUF sb;
/*void mixer ( void * userdata, u8 * stream, int len );*/	/* Mixer function proto */

unsigned short soundbuffer1[SOUND_BUFFERSIZE], soundbuffer2[SOUND_BUFFERSIZE];

struct resource {
	int rr;
	int debug;
	char bank;
	char pad_sens;
	char sc;
	char autoconfig;
	char left;
	char right;
	char swap;
} app_data;

typedef struct {
	char text[39];	/* Text to display */
/*	void (*func)(void);*/
	long func;	/* Function Number */
	unsigned char col;	/* Text color */
	char selectable;	/* Selectable item or not (0=no, 1=yes) */
	char selected;	/* Default item (0=no, 1=yes) */
} menustruct;

short MainMenuStart = 0;
#define MainMenuBack 0;
#define MainMenuSize 6
menustruct MainMenu[MainMenuSize] = {	"Main Menu", 0, MENUTEXT_COL_HEADER, 0, 0,
										"", 0, MENUTEXT_COL_NORMAL, 0, 0,
										"Select file", MENU_FILE, MENUTEXT_COL_NORMAL, 1, 1,
										"Settings", MENU_SETTINGS, MENUTEXT_COL_NORMAL, 1, 0,
										"Help", MENU_HELP, MENUTEXT_COL_NORMAL, 1, 0,
										"About", MENU_ABOUT, MENUTEXT_COL_NORMAL, 1, 0
};

short AboutMenuStart = 0;
#define AboutMenuBack MENU_MAIN;
#define AboutMenuSize 15
menustruct AboutMenu[AboutMenuSize] = {	"About VCS32 v0.1", 0, MENUTEXT_COL_HEADER, 0, 0,
										"", 0, MENUTEXT_COL_NORMAL, 0, 0,
										"VCS32 by Fabrizio Zavagli is a GP32", MENU_SETTINGS, MENUTEXT_COL_NORMAL, 0, 0,
										"porting of Virtual VCS by Dan Boris.", MENU_SETTINGS, MENUTEXT_COL_NORMAL, 0, 0,
										"Released under the GPL.", MENU_SETTINGS, MENUTEXT_COL_NORMAL, 0, 0,
										"TIA sound emulator by Ron Fries.", MENU_SETTINGS, MENUTEXT_COL_NORMAL, 0, 0,
										"In cooperation with Mike Green.", MENU_SETTINGS, MENUTEXT_COL_NORMAL, 0, 0,
										"", 0, MENUTEXT_COL_NORMAL, 0, 0,
										"Visit http://www.rasty.com for the", MENU_SETTINGS, MENUTEXT_COL_NORMAL, 0, 0,
										"latest version.", MENU_SETTINGS, MENUTEXT_COL_NORMAL, 0, 0,
										"", 0, MENUTEXT_COL_NORMAL, 0, 0,
										"Thanks: Christian Nowak, Kazzuya, Frk,", 0, MENUTEXT_COL_NORMAL, 0, 0,
										"AtariAge, Voch, Jag, Cristina, Lorenzo", 0, MENUTEXT_COL_NORMAL, 0, 0,
										"", 0, MENUTEXT_COL_NORMAL, 0, 0,
										"Return to Main Menu", MENU_MAIN, MENUTEXT_COL_NORMAL, 1, 1
};

short HelpMenuStart = 0;
#define HelpMenuBack MENU_MAIN;
#define HelpMenuSize 15
menustruct HelpMenu[HelpMenuSize] = {	"Help", 0, MENUTEXT_COL_HEADER, 0, 0,
										"", 0, MENUTEXT_COL_NORMAL, 0, 0,
										"Game:", 0, MENUTEXT_COL_NORMAL, 0, 0,
										"A = Fire", MENU_SETTINGS, MENUTEXT_COL_NORMAL, 0, 0,
										"Start = Reset", MENU_SETTINGS, MENUTEXT_COL_NORMAL, 0, 0,
										"Select = Select", MENU_SETTINGS, MENUTEXT_COL_NORMAL, 0, 0,
										"L = Set diff. switch Player 1", MENU_SETTINGS, MENUTEXT_COL_NORMAL, 0, 0,
										"R = Set diff. switch Player 2", MENU_SETTINGS, MENUTEXT_COL_NORMAL, 0, 0,
										"L+R = Back to menu", MENU_SETTINGS, MENUTEXT_COL_NORMAL, 0, 0,
										"Keep B pressed to control Player 2", MENU_SETTINGS, MENUTEXT_COL_NORMAL, 0, 0,
										"", 0, MENUTEXT_COL_NORMAL, 0, 0,
										"Menu:", 0, MENUTEXT_COL_NORMAL, 0, 0,
										"A = Select", 0, MENUTEXT_COL_NORMAL, 0, 0,
										"B = Back", 0, MENUTEXT_COL_NORMAL, 0, 0,
										"Return to Main Menu", MENU_MAIN, MENUTEXT_COL_NORMAL, 1, 1
};

short SettingsMenuStart = 0;
#define SettingsMenuBack MENU_MAIN;
#define SettingsMenuSize 5
menustruct SettingsMenu[SettingsMenuSize] = {	"Settings", 0, MENUTEXT_COL_HEADER, 0, 0,
										"", 0, MENUTEXT_COL_NORMAL, 0, 0,
										"Frameskip", MENU_SETTINGSFRAMESKIP, MENUTEXT_COL_NORMAL, 1, 1,
/*										"Brightness", MENU_SETTINGSBRIGHTNESS, MENUTEXT_COL_NORMAL, 1, 0,*/
										"", MENU_HELP, MENUTEXT_COL_NORMAL, 0, 0,
										"Back", MENU_MAIN, MENUTEXT_COL_NORMAL, 1, 0
};

short SettingsFrameskipMenuStart = 0;
#define SettingsFrameskipMenuBack MENU_SETTINGS;
#define SettingsFrameskipMenuSize 9
menustruct SettingsFrameskipMenu[SettingsFrameskipMenuSize] = {	"Frameskip", 0, MENUTEXT_COL_HEADER, 0, 0,
										"", 0, MENUTEXT_COL_NORMAL, 0, 0,
										"0", MENU_SETTINGSFRAMESKIP0, MENUTEXT_COL_NORMAL, 1, 1,
										"1", MENU_SETTINGSFRAMESKIP1, MENUTEXT_COL_NORMAL, 1, 0,
										"2 (default)", MENU_SETTINGSFRAMESKIP2, MENUTEXT_COL_NORMAL, 1, 0,
										"3", MENU_SETTINGSFRAMESKIP3, MENUTEXT_COL_NORMAL, 1, 0,
										"4", MENU_SETTINGSFRAMESKIP4, MENUTEXT_COL_NORMAL, 1, 0,
										"", MENU_HELP, MENUTEXT_COL_NORMAL, 0, 0,
										"Back", MENU_SETTINGS, MENUTEXT_COL_NORMAL, 1, 0
};

short FileMenuStart = 0;
#define FileMenuBack MENU_MAIN;
short FileMenuSize = 2;
menustruct FileMenu[2000] = {	"Select file:", 0, MENUTEXT_COL_HEADER, 0, 0,
										"", 0, MENUTEXT_COL_NORMAL, 0, 0
};

/* The mainloop from cpu.c */
/*extern void mainloop ();*/

static BOOL FileIsDir ( char * fname )
{
    GPFILEATTR attr;

    if (GpFileAttr ( fname, &attr )!=SM_OK)
      return FALSE;

    return ((attr.attr&0x10)==0x10);
}

static void sortDirEntries ( FINFO ** pdirlist, int num )
{
    int i,j;

    if (num<2)
      return;

    for (i=0;i<num;i++)
      {
        for (j=0;j<num-1;j++)
          {
            if (strcmp(pdirlist[j]->name, pdirlist[j+1]->name)>0)
              {
                FINFO * temp;
                temp = pdirlist[j];
                pdirlist[j] = pdirlist[j+1];
                pdirlist[j+1] = temp;
              }
          }
      }
}

void FPSTimer(void)
{
	/* Display number of frames drawn */
	sprintf((char*)FPS_string, "FPS: %d", frames_count);
	GpRectFill(NULL, &gpDraw[0], 0, 220, 100, 20, 0x00);
	GpTextOut(NULL, &gpDraw[0], 0, 220, (char*)FPS_string, 100);
	GpRectFill(NULL, &gpDraw[1], 0, 220, 100, 20, 0x00);
	GpTextOut(NULL, &gpDraw[1], 0, 220, (char*)FPS_string, 100);
	frames_count = 0;
}

void mixer ( void * userdata, u8 * stream, int len )
{
	Tia_process_2(stream,len);
}

void StartEmuSound()
{
	/*GpPcmInit(PCM_M11, PCM_16BIT);*/		/* Mono, 11025 hz */
	Tia_sound_init(31400,22050);
   	sb.freq = PCM_M22;
	sb.format = PCM_8BIT;
	sb.samples = 22050/60;
	sb.userdata = NULL;
	sb.callback = mixer;
	sb.pollfreq = 60; /* 12*(44100/sb.samples); */
	GpPcmInit(PCM_M22,PCM_8BIT);
	GpSoundBufStart(&sb);
}

void StopEmuSound()
{
	GpSoundBufStop();
}

void GpInitStuff()
{
	int i, c;
	/*GpClockSpeedChange(132000000, 0x24001, 2);*/ /* o */
    GpClockSpeedChange (132000000, 0x3a011, 3);
	/*GpClockSpeedChange(160000000, 0x24001, 2);*/

	/*GpGraphicModeSet(GPC_DFLAG_16BPP,NULL);*/
	/*GpGraphicModeSet(GPC_DFLAG_8BPP,NULL);*/

	GpFatInit();

	/*StartEmuSound();*/

	vidbuf.buf_w = 320 & ~3;	/*should be 4-byte aligned!!!*/
	vidbuf.buf_h = 240 & ~3;	/*should be 4-byte aligned!!!*/
	GpMemSurfaceGet(&vidbuf);

	/*turn on the gp32 lcd screen*/
	for(i = 0 ; i < 2 ; i++)
	{
		GpLcdSurfaceGet(&gpDraw[i], i); 
	} 
	/*clear it*/
	GpRectFill(NULL, &gpDraw[0], 0, 0, gpDraw[0].buf_w, gpDraw[0].buf_h, 0x00);
	GpRectFill(NULL, &gpDraw[1], 0, 0, gpDraw[1].buf_w, gpDraw[1].buf_h, 0x00);

/*
	for (c=0; c<320*240; c++) {
		vidbuf.ptbuffer[c] = 0xe0;
		vidbuf.ptbuffer[c] = c & 254;
	}
*/
	/*set the pallete, which is also in the file generated by gp32 converter*/
	/*
	h_pal = GpPaletteCreate(256, (GP_PALETTEENTRY*)background_Pal);
		GpPaletteSelect(h_pal);
		GpPaletteRealize();
		GpPaletteDelete(h_pal);
		h_pal = NULL;	
	*/

	/*set the serface we want to be shown on the gp32 screen*/
	GpSurfaceSet(&gpDraw[0]);

	#ifdef SW_FPSCOUNT
	/* TIMER */
	
	frames_count = 0;
	if(GPOS_ERR_ALREADY_USED == GpTimerOptSet(0,1,0,FPSTimer))	GpTimerKill(0);
	GpTimerSet(0);
	
	#endif
}

void initmenu()
{

	h_pal_menu = GpPaletteCreate(256, (GP_PALETTEENTRY*)palette_menu);
	GpPaletteSelect(h_pal_menu);
	GpPaletteRealize();
	/*GpPaletteDelete(h_pal_menu);
	h_pal_menu = NULL;*/

}

void initmenu_full()
{
	/* Init the Menu Palette */
	/*unsigned short rgb;*/
	unsigned long rgb;
	short i, c;
	BYTE red,green,blue;

	c=0;
	for(i=0; i<256; i++)
	{
		red  = (BYTE)menubg_pal[c++];
		green  = (BYTE)menubg_pal[c++];
		blue  = (BYTE)menubg_pal[c++];
		/* Converting to 5:5:5:1 */
		rgb = 0;
		rgb = (red >> 3) << 11;
		rgb |= (green >> 3) << 6;
		rgb |= (blue >> 3) << 1;
		palette_menu[i] = rgb;
	}
	initmenu();
}

short DisplayMenu(menustruct *Menu, short menusize, short *menustart)
{
	short i, xoff, y;
	unsigned char col;
	short menusel, c;

	c = 0;	/* Cursor position on screen line */
	i=*menustart;
	while ((c<menusize) && (c<MENU_SCREENSIZE)) {
		xoff = 0;
		col = Menu[i].col;
		if (Menu[i].selected) {
			col = MENUTEXT_COL_SELECT;
			xoff = 5;
			menusel = i;
			menu_cursor = c;
		}
		GpTextOut(NULL, &vidbuf.ptbuffer, xoff+MENUOFFSET_X, (c*MENUKERNING)+MENUOFFSET_Y, Menu[i].text, col);
		c++;
		i++;
	}
	return menusel;
}

void domenu()
{
	/*GpRectFill(NULL, &vidbuf.ptbuffer, 0, 0, vidbuf.buf_w, vidbuf.buf_h, 50);*/
	short menusel, menusize, c, *menustart, menuback;
	menustruct *Menu;
	long movedelay=100000;

	StopEmuSound();

#ifdef RELEASE
	do {	/* Enable this to check for rom loaded 1st time */
#endif
	do {
	/*while (!((ExKey = GpKeyGet()) & GPC_VK_FB)) {*/
		ExKey = GpKeyGet();
		/* Blit background */
		GpBitBlt(NULL,
					&vidbuf.ptbuffer,
					0, 0,
					320, 240,
					menubg,
					0, 0,
					320, 240);

		/* Display Selected Menu */
		/*
		GpTextOut(NULL, &vidbuf.ptbuffer, 0+MENUOFFSET_X, 0+MENUOFFSET_Y, "Menu", MENUTEXT_COL_HEADER);
		GpTextOut(NULL, &vidbuf.ptbuffer, 0+MENUOFFSET_X, 15+MENUOFFSET_Y, filebuffer, MENUTEXT_COL_HEADER);
		*/
		switch (current_menu)
		{
			case MENU_MAIN: {
				menusize = MainMenuSize;
				Menu = &MainMenu;
				menustart = &MainMenuStart;
				menuback = MainMenuBack;
				break;
			}
			case MENU_HELP: {
				menusize = HelpMenuSize;
				Menu = &HelpMenu;
				menustart = &HelpMenuStart;
				menuback = HelpMenuBack;
				break;
			}
			case MENU_ABOUT: {
				menusize = AboutMenuSize;
				Menu = &AboutMenu;
				menustart = &AboutMenuStart;
				menuback = AboutMenuBack;
				break;
			}
			case MENU_SETTINGS: {
				menusize = SettingsMenuSize;
				Menu = &SettingsMenu;
				menustart = &SettingsMenuStart;
				menuback = SettingsMenuBack;
				break;
			}
			case MENU_SETTINGSFRAMESKIP: {
				menusize = SettingsFrameskipMenuSize;
				Menu = &SettingsFrameskipMenu;
				menustart = &SettingsFrameskipMenuStart;
				menuback = SettingsFrameskipMenuBack;
				break;
			}
			case MENU_FILE: {
				menusize = FileMenuSize;
				Menu = &FileMenu;
				menustart = &FileMenuStart;
				menuback = FileMenuBack;
				break;
			}
			case MENU_SETTINGSFRAMESKIP0: {
				app_data.rr = 1;
				current_menu = MENU_SETTINGS;
				break;
			}
			case MENU_SETTINGSFRAMESKIP1: {
				app_data.rr = 2;
				current_menu = MENU_SETTINGS;
				break;
			}
			case MENU_SETTINGSFRAMESKIP2: {
				app_data.rr = 3;
				current_menu = MENU_SETTINGS;
				break;
			}
			case MENU_SETTINGSFRAMESKIP3: {
				app_data.rr = 4;
				current_menu = MENU_SETTINGS;
				break;
			}
			case MENU_SETTINGSFRAMESKIP4: {
				app_data.rr = 5;
				current_menu = MENU_SETTINGS;
				break;
			}
		}

		menusel = DisplayMenu(Menu, menusize, menustart);

		/* Check joystick */
		if (!movedelay) {
			if (ExKey & GPC_VK_FB) {
				current_menu = menuback;
				movedelay = 100000;
			}
			if (ExKey & GPC_VK_FA) {
				/* File selector invoked? */
				if ((Menu[menusel].func != current_menu) || (Menu[menusel].func >= 1000)) {
					current_menu = Menu[menusel].func;
					movedelay = MENU_MOVEDELAY;
					/* Initialize the menu (if needed) */
					if (current_menu >= 1000) {	/* A file has been selected! */
						/*strcpy(menu_filename, Menu[current_menu-1000].text);*/
						loadNewCart(Menu[current_menu-1000].text);
						mainloop_init();
						current_menu = MENU_EMU;
					}
				}
			}
			if ((ExKey & GPC_VK_UP) || (ExKey & GPC_VK_FL)) {
				c = -1;
				while ((!Menu[menusel+c].selectable) && (menusel+c >= 0)) {
					c--;
				}
				if (((menusel+c) >= 0)) { /*&& Menu[menusel+1].selectable) {*/
					Menu[menusel].selected = 0;
					Menu[menusel+c].selected = 1;
					if (ExKey & GPC_VK_UP) movedelay = MENU_MOVEDELAY;
					else movedelay = MENU_MOVEDELAY_FAST;
					if (menu_cursor+c < 0) *menustart += c;
				}
			}
			if ((ExKey & GPC_VK_DOWN) || (ExKey & GPC_VK_FR)) {
				c = 1;
				while ((!Menu[menusel+c].selectable) && (menusel+c < menusize)) {
					c++;
				}
				if (((menusel+c) < menusize)) { /*&& Menu[menusel+1].selectable) {*/
					Menu[menusel].selected = 0;
					Menu[menusel+c].selected = 1;
					if (ExKey & GPC_VK_DOWN) movedelay = MENU_MOVEDELAY;
					else movedelay = MENU_MOVEDELAY_FAST;
					if (menu_cursor+c >= MENU_SCREENSIZE) *menustart += c;
				}
			}
		} else movedelay--;

		if (!(ExKey & GPC_VK_DOWN) && !(ExKey & GPC_VK_UP) && !(ExKey & GPC_VK_FR) && !(ExKey & GPC_VK_FL) && !(ExKey & GPC_VK_FA) && !(ExKey & GPC_VK_FB)) {
			movedelay = 0;
		}

		/* Blit vbuffer into screen buffer */
		GpBitBlt(NULL,
					&gpDraw[nflip],
					0, 0,
					320, 240,
					vidbuf.ptbuffer,
					0, 0,
					320, 240);
		GpSurfaceFlip(&gpDraw[nflip++]);	/* Flips buffers and waits for beam */
		nflip &= 0x01;
	} while (current_menu);
#ifdef RELEASE
	} while (!cart_loaded);
#endif

/*	ExKey = GpKeyGet();*/
	current_screen = SCREEN_EMU;	/* Resume emulation */
	StartEmuSound();
	GpPaletteSelect(h_pal);
	GpPaletteRealize();
	GpRectFill(NULL, &gpDraw[0], 0, 0, gpDraw[0].buf_w, gpDraw[0].buf_h, 0x00);
	GpRectFill(NULL, &gpDraw[1], 0, 0, gpDraw[1].buf_w, gpDraw[1].buf_h, 0x00);
}

static int SelectFile ( FSEL * fsel, char * basedir)
  {
    unsigned int i;
    unsigned int j;
    int numdirs = 0;
    ulong read_count;
    int index = 0;
    int lk = 0,k;
    FINFO * dirlist;
    GPDIRENTRY * direntry;
    char p[256];

    /* Get directory */
	num_entries = 0;
    if ( GpDirEnumNum(basedir, &num_entries)!=SM_OK )
      return -1;
    if (num_entries==0)
      return -1;

    dirlist = malloc ( num_entries*sizeof(FINFO) );
    memset ( dirlist, 0, sizeof(FINFO)*num_entries );
    pdirlist = malloc ( num_entries*sizeof(FINFO*) );

    direntry = malloc ( sizeof(GPDIRENTRY)*num_entries );
    GpDirEnumList ( basedir, 0, num_entries, direntry, &read_count );
    num_entries = read_count;

    for (i=0;i<num_entries;i++)
      {
        strcpy ( dirlist[i].name, direntry[i].name );
        if ( (strcmp("." , direntry[i].name)==0) ||
             (strcmp("..", direntry[i].name)==0) ||
             FileIsDir(direntry[i].name) )
          {
            dirlist[i].isdir = TRUE;
            numdirs++;
            dirlist[i].size = -1;
          }
        else
          {
            ulong s = 0;
            GpFileGetSize ( direntry[i].name, &s );
            dirlist[i].isdir = FALSE;
            dirlist[i].size = (int)s;
          }
        pdirlist[i] = &dirlist[i];
      }

    free ( direntry );

    /* Separate dirs from files */
    j=0;i=0;
    while (pdirlist[j]->isdir && j<num_entries) j++;
    for (i=j;i<num_entries;i++)
      {
        if (pdirlist[i]->isdir)
          {
            FINFO * tmp = pdirlist[i];
            pdirlist[i] = pdirlist[j];
            pdirlist[j] = tmp;
            j++;
          }
      }
    /* Sort directories */
    sortDirEntries ( pdirlist, numdirs );
    /* Sort files */
    sortDirEntries ( &pdirlist[numdirs], num_entries-numdirs );

	/*
    DisplayEntries ( fsel, pdirlist, num_entries, index );
    GpRelativePathGet(p);
    GpTextOut(NULL,fsel->gpDraw,0,0,p,fsel->curdircolor);
    while (GpKeyGet()&GPC_VK_FA);
    while (!((k=GpKeyGet())&(GPC_VK_FA|(enableCancel?GPC_VK_FB:0))) )
      {
        k &= GPC_VK_UP | GPC_VK_DOWN | GPC_VK_LEFT | GPC_VK_RIGHT;

        if (lk==0)
          {
            if (k&GPC_VK_UP)
              index--;
            else
            if (k&GPC_VK_DOWN)
              index++;
            else
            if (k&GPC_VK_LEFT)
              index-=15;
            else
            if (k&GPC_VK_RIGHT)
              index+=15;

            if (index<0) index=0;
            if (index>=(int)num_entries) index=num_entries-1;

            Display entries
            if (k&(GPC_VK_UP|GPC_VK_DOWN|GPC_VK_LEFT|GPC_VK_RIGHT))
              {
                DisplayEntries ( fsel, pdirlist, num_entries, index );
                GpRelativePathGet(p);
                GpTextOut(NULL,fsel->gpDraw,0,0,p,fsel->curdircolor);
              }
          }

        lk = k;
      }
	strcpy(fsel->fname, pdirlist[index]->name);
    free ( pdirlist );
    free ( dirlist );

    if ( enableCancel )
      return k&GPC_VK_FB ? -1 : 0;
    else
      return 0;
	*/

	FileMenu[FileMenuSize].selected = 1;
	for (i=numdirs; i<num_entries; i++) {
		strcpy(FileMenu[FileMenuSize].text, pdirlist[i]->name);
		/*sprintf(FileMenu[FileMenuSize].text, "%s - %d", pdirlist[i]->name, FileMenu[FileMenuSize].func);*/
		FileMenu[FileMenuSize].func = i+1000;
		FileMenu[FileMenuSize].col = MENUTEXT_COL_NORMAL;
		FileMenu[FileMenuSize].selectable = 1;
		if (i>numdirs) FileMenu[FileMenuSize].selected = 0;
		FileMenuSize++;
	}
}

void InitFileMenu(FSEL *fsel, char *bdir)
  {
    char * basedir;
    BOOL isdir = FALSE;

    basedir=bdir;

	if (!FileIsDir(basedir))
      {
        basedir="gp:\\gpmm";
        if (!FileIsDir(basedir))
          {
            basedir="gp:\\";
          }
      }

	GpRelativePathSet ( basedir );

    do
      {
        char path[256];
        if (SelectFile(fsel, "\\")!=0)
          return -1;
        GpRelativePathGet(path);
        if (strcmp(fsel->fname,"..")==0)
          {
            int i;
            int l=strlen(path);
            path[l-1]='0';
            for (i=l-2;i>0 && path[i]!='\\';i--)
              path[i]='\0';
            isdir=TRUE;
          } else
        if (strcmp(fsel->fname,".")==0)
          {
            isdir=TRUE;
          } else
          {
            if ((isdir=FileIsDir(fsel->fname)))
              strcat(path,fsel->fname);
          }

        if (isdir)
          GpRelativePathSet(path);
        GpRelativePathGet(path);
      } while (isdir);
    return 0;
  }

void GpMain(void * arg)
{
	BYTE b;
	unsigned long c;
	unsigned short xx=0, yy=0;

	int i;
	char file[50];
	char attr[50],val[50];
	char *p;

	app_data.rr=1;
	app_data.debug=0;
	app_data.bank=2;
	app_data.pad_sens=5;
	app_data.sc=0;
	app_data.left=STICK;
	app_data.right=STICK;
/*	paddle1.pos=0;
	paddle1.val=0;*/
	app_data.autoconfig=0;
	app_data.swap=0;
	app_data.rr = 3;	/* Frameskip */

	/* load image */
	loadCart();

	GpInitStuff();	/* Init GP32-specific stuff */
	/*for (i=0; i<256; i++) keys[i] = 0;*/
	/*Tia_sound_init(31400, 11025);*/

	mainloop_init();
	initmenu_full();
	InitFileMenu(&filesel, VCSbasedir);

	current_screen = SCREEN_MENU;
	/*current_screen = SCREEN_EMU;*/
	current_menu = MENU_MAIN;

	nflip=1;
	c=0;
	filesel.gpDraw = &gpDraw[0];
    filesel.bgcolor = 0;       /* Background color */
    filesel.dircolor = 100;      /* Directory entry color */
    filesel.filecolor = 110;     /* File entry color */
    filesel.dirselcolor = 90;   /* Selected directory entry color */
    filesel.fileselcolor = 90;  /* Selected file entry color */
    filesel.curdircolor = 50;   /* Color of current directory display at the top of the screen */
	filesel.fname = &filebuffer[0];
	/*FileRequest ( &filesel, "gp:\\gpmm", 0 );*/
	while (1) {	

	/* Draw here on vidbuf.ptbuffer[(x*240 + 239-y)] = color from palette */
	/*vidbuf.ptbuffer[0*240 + 239-0] = 0;*/
	/*gpDraw[nflip].ptbuffer[0*240 + 239-0] = 0;*/
/*
	if (xx<319) xx++;
	if (yy<239) yy++;
	gpDraw[nflip].ptbuffer[xx*240 + 239-yy] = 0;
*/

	/* start cpu */
	switch (current_screen) {
		case SCREEN_MENU: {
			domenu();
		}
		break;
		case SCREEN_EMU: {
			mainloop();
			current_menu = MENU_MAIN;
			initmenu();
		}
		break;
	}

	/*memcpy(&vidbuf.ptbuffer[0],&vscreen[0],64000);*/
	/*memcpy(&vidbuf.ptbuffer[0],&theRom[0],4096);*/
/*	
	for (i=0; i<4096; i++) {
		vidbuf.ptbuffer[i] = theRom[i];
	}
*/	
	/*GpTextOut(NULL, &gpDraw[nflip], 0, 0, "prova", 1); */

	/*see what buttons are being pressed*/
	/*
	ExKey = GpKeyGet();
	if (bulletdelay<10) bulletdelay++;
	if (ExKey & GPC_VK_FA) {if (bulletdelay==10) {bul++, bulx[bul]=sx+28,buly[bul]=sy+12; bulletdelay=0;}}
	if (bul>13) bul=0;
	if (ExKey & GPC_VK_LEFT) if (sx>5) sx-=2;
	if (ExKey & GPC_VK_RIGHT) if (sx<300) sx+=2;
	if (ExKey & GPC_VK_UP) if (sy>5) sy-=2;
	if (ExKey & GPC_VK_DOWN) if (sy<164) sy+=2;
	if (ExKey & GPC_VK_START ) GpAppExit();
	*/

	/* Copy offscreen buffer to screen buffer */
	/*
	GpBitBlt(NULL,
				&gpDraw[nflip],
				0, 0,
				320, 240,
				vidbuf.ptbuffer,
				0, 0,
				320, 240);
	*/
	/*flip the lcd surcace over, makes it very smooth as we are always drawing on the
	one no one can see then flipping it over when its finished*/
	/* Flips buffers and waits for beam */
	/*	GpSurfaceFlip(&gpDraw[nflip++]);
	nflip &= 0x01;*/
	}
}
